#!/usr/bin/env bash
set -o pipefail
declare -r PLOWSHARE_API_VERSION=6
declare -r PLOWSHARE_CONFDIR="${XDG_CONFIG_HOME:-$HOME/.config}/plowshare"
declare -r PLOWCORE_JS=${PLOWSHARE_JS:-js}
declare -r PLOWCORE_DEBIAN_JS=${PLOWSHARE_DEBIAN_JS:-no}
declare -r PLOWCORE_CURL=${PLOWSHARE_CURL:-curl}
declare -r ERR_FATAL=1                    # Unexpected result (upstream site updated, etc)
declare -r ERR_NOMODULE=2                 # No module found for processing request
declare -r ERR_NETWORK=3                  # Generic network error (socket failure, curl, firewall, etc)
declare -r ERR_LOGIN_FAILED=4             # Correct login/password argument is required
declare -r ERR_MAX_WAIT_REACHED=5         # Wait timeout (see -t/--timeout command line option)
declare -r ERR_MAX_TRIES_REACHED=6        # Max tries reached (see -r/--max-retries command line option)
declare -r ERR_CAPTCHA=7                  # Captcha solving failure
declare -r ERR_SYSTEM=8                   # System failure (missing executable, local filesystem, wrong behavior, etc)
declare -r ERR_LINK_TEMP_UNAVAILABLE=10   # plowdown: Link alive but temporarily unavailable
declare -r ERR_LINK_PASSWORD_REQUIRED=11  # plowdown: Link alive but requires a password
declare -r ERR_LINK_NEED_PERMISSIONS=12   # plowdown: Link alive but requires some authentication (private or premium link)
declare -r ERR_LINK_DEAD=13               # plowdel: File not found or previously deleted
declare -r ERR_SIZE_LIMIT_EXCEEDED=14     # plowdown: Can't download link because file is too big (need permissions)
declare -r ERR_BAD_COMMAND_LINE=15        # Unknown command line parameter or incompatible options
declare -r ERR_ASYNC_REQUEST=16           # plowup: Asynchronous remote upload started (can't predict final status)
declare -r ERR_EXPIRED_SESSION=17         # Related to local storage module file, expired session
declare -r ERR_FATAL_MULTIPLE=100         # 100 + (n) with n = first error code (when multiple arguments)
logcat_report() {
 if test -s "$1"; then
  test $VERBOSE -lt 4 || \
   stderr "$(sed -e 's/^/rep:/' "$1")"
 fi
 return 0
}
log_report() {
 test $VERBOSE -lt 4 || stderr 'rep:' "$@"
}
log_debug() {
 test $VERBOSE -lt 3 || stderr 'dbg:' "$@"
}
log_notice() {
 if [[ $COLOR ]]; then
  test $VERBOSE -lt 2 || echo -e "\033[0;33m$@\033[0m" >&2
 else
  test $VERBOSE -lt 2 || stderr "$@"
 fi
}
log_error() {
 if [[ $COLOR ]]; then
  test $VERBOSE -lt 1 || echo -e "\033[0;31m$@\033[0m" >&2
 else
  test $VERBOSE -lt 1 || stderr "$@"
 fi
}
curl() {
 local -a CURL_ARGS=("$@")
 local -a OPTIONS=(--insecure --compressed --speed-time 600 --connect-timeout 240)
 local DRETVAL=0
 if ! find_in_array CURL_ARGS[@] '-A' '--user-agent'; then
  OPTIONS+=(--user-agent \
   'Mozilla/5.0 (X11; Linux x86_64; rv:6.0) Gecko/20100101 Firefox/6.0')
 fi
 if find_in_array CURL_ARGS[@] '-L' '--location'; then
  find_in_array CURL_ARGS[@] '--max-redirs' || OPTIONS+=(--max-redirs 5)
 fi
 if [ -n "$NO_CURLRC" ]; then
  OPTIONS=(-q "${OPTIONS[@]}")
 elif [ -n "$EXT_CURLRC" ]; then
  OPTIONS=(-q --config "$EXT_CURLRC" "${OPTIONS[@]}")
 fi
 if [ "${FUNCNAME[1]}" = 'curl_with_log' ]; then
  test $VERBOSE -eq 0 && OPTIONS[${#OPTIONS[@]}]='--silent'
 else
  test $VERBOSE -ne 3 && OPTIONS[${#OPTIONS[@]}]='--silent'
 fi
 if test -n "$INTERFACE"; then
  OPTIONS+=(--interface $INTERFACE)
 fi
 if test -n "$MAX_LIMIT_RATE"; then
  OPTIONS+=(--limit-rate $MAX_LIMIT_RATE)
 fi
 if test -n "$MIN_LIMIT_RATE"; then
  OPTIONS+=(--speed-time 30 --speed-limit $MIN_LIMIT_RATE)
 fi
 if test $VERBOSE -lt 4; then
  command "$PLOWCORE_CURL" "${OPTIONS[@]}" "${CURL_ARGS[@]}" || DRETVAL=$?
 else
  local FILESIZE TEMPCURL=$(create_tempfile)
  log_report "${OPTIONS[@]}" "${CURL_ARGS[@]}"
  command "$PLOWCORE_CURL" "${OPTIONS[@]}" "${CURL_ARGS[@]}" --show-error --silent >"$TEMPCURL" 2>&1 || DRETVAL=$?
  FILESIZE=$(get_filesize "$TEMPCURL")
  log_report "Received $FILESIZE bytes. DRETVAL=$DRETVAL"
  log_report '=== CURL BEGIN ==='
  logcat_report "$TEMPCURL"
  log_report '=== CURL END ==='
  cat "$TEMPCURL"
  rm -f "$TEMPCURL"
 fi
 if [ "$DRETVAL" != 0 ]; then
  local INDEX FILE
  if INDEX=$(index_in_array CURL_ARGS[@] '-D' '--dump-header'); then
   FILE=${OPTIONS[$INDEX]}
   if [ -f "$FILE" ]; then
    log_debug "deleting temporary HTTP header file: $FILE"
    rm -f "$FILE"
   fi
  fi
  if INDEX=$(index_in_array CURL_ARGS[@] '-o' '--output'); then
   FILE=${OPTIONS[$INDEX]}
   if [ -f "$FILE" ] && ! find_in_array OPTIONS[@] '--globoff'; then
    log_debug "deleting temporary output file: $FILE"
    rm -f "$FILE"
   fi
  fi
  case $DRETVAL in
   2|27)
    log_error "$FUNCNAME: out of memory?"
    return $ERR_SYSTEM
    ;;
   6)
    log_notice "$FUNCNAME: couldn't resolve host"
    return $ERR_NETWORK
    ;;
   7)
    log_notice "$FUNCNAME: couldn't connect to host"
    return $ERR_NETWORK
    ;;
   18)
    return $ERR_LINK_TEMP_UNAVAILABLE
    ;;
   22|28)
    log_error "$FUNCNAME: HTTP retrieve error"
    return $ERR_NETWORK
    ;;
   23)
    log_error "$FUNCNAME: write failed, disk full?"
    return $ERR_SYSTEM
    ;;
   47)
    if ! find_in_array CURL_ARGS[@] '--max-redirs'; then
     log_error "$FUNCNAME: too many redirects"
     return $ERR_FATAL
    fi
    ;;
   62)
    return $ERR_SYSTEM
    ;;
   *)
    log_error "$FUNCNAME: failed with exit code $DRETVAL"
    return $ERR_NETWORK
    ;;
  esac
 fi
 return 0
}
curl_with_log() {
 curl "$@"
}
replace_all() {
 local S=$(cat)
 local FROM=${1//\\/\\\\}
 echo "${S//$FROM/$2}"
}
replace() {
 local S=$(cat)
 local FROM=${1//\\/\\\\}
 echo "${S/$FROM/$2}"
}
uppercase() {
 echo "${*^^}"
}
lowercase() {
 echo "${*,,}"
}
first_line() {
 local -r N=${1:-1}
 if (( N < 1 )); then
  log_error "$FUNCNAME: negative index not expected"
  return $ERR_FATAL
 fi
 head -n$((N))
}
last_line() {
 local -r N=${1:-1}
 if (( N < 1 )); then
  log_error "$FUNCNAME: negative index not expected"
  return $ERR_FATAL
 fi
 tail -n$((N))
}
nth_line() {
 local -r N=${1:-1}
 if (( N < 1 )); then
  log_error "$FUNCNAME: negative index not expected"
  return $ERR_FATAL
 fi
 sed -ne "$((N))p"
}
delete_first_line() {
 local -r N=${1:-1}
 if (( N < 1 )); then
  log_error "$FUNCNAME: negative index not expected"
  return $ERR_FATAL
 fi
 sed -ne "$((N+1)),\$p"
}
delete_last_line() {
 local -r N=${1:-1}
 if (( N < 1 )); then
  log_error "$FUNCNAME: negative index not expected"
  return $ERR_FATAL
 fi
 sed -ne :a -e "1,$N!{P;N;D;};N;ba"
}
delete_filter_line() {
 local -i FUZZ=${2:-0}
 local -i N=${3:-1}
 local -r D=$'\001' # Change sed separator to allow '/' characters in regexp
 local STR FILTER
 if [[ ! $1 ]]; then
  log_error "$FUNCNAME: invalid regexp, must not be empty"
  return $ERR_FATAL
 elif [ $N -le 0 ]; then
  log_error "$FUNCNAME: wrong argument, start line must be strictly positive ($N given)"
  return $ERR_FATAL
 elif [ $N -gt 1 -a $FUZZ -lt -1 ]; then
  log_debug "$FUNCNAME: before context ($FUZZ) could duplicates lines (already printed before line $N), continue anyway"
 fi
 (( --N ))
 FILTER="\\${D}$1${D}" # /$1/
 if [ $FUZZ -eq 0 ]; then
  if (( $N > 0 )); then
   STR=$(sed -e "${N}p" -e "$N,${FILTER}d")
  else
   STR=$(sed -e "$N,${FILTER}d")
  fi
 elif [ $FUZZ -eq 1 ]; then
  if (( $N > 0 )); then
   STR=$(sed -e "${N}p" -e "$N,${FILTER}{${FILTER}N;d}")
  else
   STR=$(sed -e "$N,${FILTER}{${FILTER}N;d}")
  fi
 elif [ $FUZZ -eq -1 ]; then
  if (( $N > 0 )); then
   STR=$(sed -e "${N}p" -e "$N,${FILTER}{${N}d;${FILTER}p;d}")
  else
   STR=$(sed -e "$N,${FILTER}{${FILTER}p;d}")
  fi
 else
  local -i FUZZ_ABS=$(( FUZZ < 0 ? -FUZZ : FUZZ ))
  [ $FUZZ_ABS -gt 10 ] &&
   log_notice "$FUNCNAME: are you sure you want to skip $((N+1)) lines?"
  if [ $FUZZ -gt 0 ]; then
   local SKIPS='N'
   while (( --FUZZ_ABS )); do
    SKIPS+=';N'
   done
   if (( $N > 0 )); then
    STR=$(sed -e "${N}p" -e "$N,${FILTER}{${FILTER}{${SKIPS}};d}")
   else
    STR=$(sed -e "$N,${FILTER}{${FILTER}{${SKIPS}};d}")
   fi
  else
   local LINES='.*'
   while (( --FUZZ_ABS )); do
    LINES+='\n.*'
   done
   if (( $N > 0 )); then
    STR=$(sed -e "${N}p" -e "1h;1!H;x;s/^.*\\n\\($LINES\)$/\\1/;x" \
        -e "$N,${FILTER}{${N}d;${FILTER}{g;p};d}")
   else
    STR=$(sed -e "1h;1!H;x;s/^.*\\n\\($LINES\)$/\\1/;x" \
        -e "$N,${FILTER}{${FILTER}{g;p};d}")
   fi
  fi
 fi
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"$N,/$1/d\" (skip $FUZZ)"
  log_notice_stack
  return $ERR_FATAL
 fi
 echo "$STR"
}
match() {
 if [ -z "$2" ]; then
  log_debug "$FUNCNAME: input buffer is empty"
  return $ERR_FATAL
 else
  grep -q -- "$1" <<< "$2"
 fi
}
matchi() {
 if [ -z "$2" ]; then
  log_debug "$FUNCNAME: input buffer is empty"
  return $ERR_FATAL
 else
  grep -iq -- "$1" <<< "$2"
 fi
}
match_remote_url() {
 local -r URL=$1
 local RET=$ERR_FATAL
 [[ $URL =~ ^[[:space:]]*[Hh][Tt][Tt][Pp][Ss]?:// ]] && return 0
 shopt -s nocasematch
 while [[ $# -gt 1 ]]; do
  shift
  if [[ $1 =~ ^[[:alpha:]][-.+[:alnum:]]*$ ]]; then
   if [[ $URL =~ ^[[:space:]]*$1:// ]]; then
    RET=0
    break
   fi
  else
   log_error "$FUNCNAME: invalid scheme \`$1'"
   break
  fi
 done
 shopt -u nocasematch
 return $RET
}
parse_all() {
 local PARSE=$2
 local -i N=${3:-0}
 local -r D=$'\001' # Change sed separator to allow '/' characters in regexp
 local STR FILTER
 if [ -n "$1" -a "$1" != '.' ]; then
  FILTER="\\${D}$1${D}" # /$1/
 elif [ $N -ne 0 ]; then
  log_error "$FUNCNAME: wrong argument, offset argument is $N and filter regexp is \"$1\""
  return $ERR_FATAL
 fi
 [ '^' = "${PARSE:0:1}" ] || PARSE="^.*$PARSE"
 [ '$' = "${PARSE:(-1):1}" ] || PARSE+='.*$'
 PARSE="s${D}$PARSE${D}\1${D}p" # s/$PARSE/\1/p
 if [ $N -eq 0 ]; then
  STR=$(sed -ne "$FILTER $PARSE")
 elif [ $N -eq 1 ]; then
  STR=$(sed -ne ":a $FILTER {n;h; $PARSE; g;ba;}")
 elif [ $N -eq -1 ]; then
  STR=$(sed -ne "$FILTER {x; $PARSE; b;}" -e 'h')
 else
  local -r FIRST_LINE='^\([^\n]*\).*$'
  local -r LAST_LINE='^.*\n\(.*\)$'
  local N_ABS=$(( N < 0 ? -N : N ))
  local I=$(( N_ABS - 2 )) # Note: N_ABS >= 2 due to "elif" above
  local LINES='.*'
  local INIT='N'
  local FILTER_LINE PARSE_LINE
  [ $N_ABS -gt 10 ] &&
   log_notice "$FUNCNAME: are you sure you want to skip $N lines?"
  while (( I-- )); do
   INIT+=';N'
  done
  while (( N_ABS-- )); do
   LINES+='\n.*'
  done
  if [ $N -gt 0 ]; then
   FILTER_LINE=$FIRST_LINE
   PARSE_LINE=$LAST_LINE
  else
   FILTER_LINE=$LAST_LINE
   PARSE_LINE=$FIRST_LINE
  fi
  STR=$(sed -ne "1 {$INIT;h;n}" \
   -e "H;g;s/^.*\\n\\($LINES\)$/\\1/;h" \
   -e "s/$FILTER_LINE/\1/" \
   -e "$FILTER {g;s/$PARSE_LINE/\1/;$PARSE }")
 fi
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"/$1/ ${PARSE//$D//}\" (skip $N)"
  log_notice_stack
  return $ERR_FATAL
 fi
 echo "$STR"
}
parse_all_quiet() {
 parse_all "$@" 2>/dev/null
 return 0
}
parse() {
 local PARSE=$2
 local -i N=${3:-0}
 local -r D=$'\001' # Change sed separator to allow '/' characters in regexp
 local STR FILTER
 if [ -n "$1" -a "$1" != '.' ]; then
  FILTER="\\${D}$1${D}" # /$1/
 elif [ $N -ne 0 ]; then
  log_error "$FUNCNAME: wrong argument, offset argument is $N and filter regexp is \"$1\""
  return $ERR_FATAL
 fi
 [ '^' = "${PARSE:0:1}" ] || PARSE="^.*$PARSE"
 [ '$' = "${PARSE:(-1):1}" ] || PARSE+='.*$'
 PARSE="s${D}$PARSE${D}\1${D}p" # s/$PARSE/\1/p
 if [ $N -eq 0 ]; then
  STR=$(sed -ne "$FILTER {$PARSE;T;q;}")
 elif [ $N -eq 1 ]; then
  STR=$(sed -ne ":a $FILTER {n;$PARSE;Ta;q;}")
 elif [ $N -eq -1 ]; then
  STR=$(sed -ne "$FILTER {g;$PARSE;T;q;}" -e 'h')
 else
  local -r FIRST_LINE='^\([^\n]*\).*$'
  local -r LAST_LINE='^.*\n\(.*\)$'
  local N_ABS=$(( N < 0 ? -N : N ))
  local I=$(( N_ABS - 2 ))
  local LINES='.*'
  local INIT='N'
  local FILTER_LINE PARSE_LINE
  [ $N_ABS -gt 10 ] &&
   log_notice "$FUNCNAME: are you sure you want to skip $N lines?"
  while (( I-- )); do
   INIT+=';N'
  done
  while (( N_ABS-- )); do
   LINES+='\n.*'
  done
  if [ $N -gt 0 ]; then
   FILTER_LINE=$FIRST_LINE
   PARSE_LINE=$LAST_LINE
  else
   FILTER_LINE=$LAST_LINE
   PARSE_LINE=$FIRST_LINE
  fi
  STR=$(sed -ne "1 {$INIT;h;n}" \
   -e "H;g;s/^.*\\n\\($LINES\)$/\\1/;h" \
   -e "s/$FILTER_LINE/\1/" \
   -e "$FILTER {g;s/$PARSE_LINE/\1/;ta;:a;$PARSE;T;q;}")
 fi
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"/$1/ ${PARSE//$D//}\" (skip $N)"
  log_notice_stack
  return $ERR_FATAL
 fi
 echo "$STR"
}
parse_quiet() {
 parse "$@" 2>/dev/null
 return 0
}
parse_json() {
 local -r NAME="\"$1\"[[:space:]]*:[[:space:]]*"
 local STR PRE
 local -r END='\([,}[:space:]].*\)\?'
 if [ "$2" = 'join' ]; then
  PRE="tr -d '\n\r'"
 elif [ "$2" = 'split' ]; then
  PRE=sed\ -e\ 's/,[[:space:]]*\(["{]\)/\n\1/g'
 else
  PRE='cat'
 fi
 STR=$($PRE | sed \
  -ne "/$NAME\[/{s/^.*$NAME\(\[[^]]*\]\).*$/\1/;ta;:a;s/^\[.*\[//;t;p;q;}" \
  -ne "/$NAME{/{s/^.*$NAME\({[^}]*}\).*$/\1/;ta;:a;s/^{.*{//;t;p;q;}" \
  -ne "s/^.*$NAME\(-\?\(0\|[1-9][[:digit:]]*\)\(\.[[:digit:]]\+\)\?\([eE][-+]\?[[:digit:]]\+\)\?\)$END/\1/p" \
  -ne "s/^.*$NAME\(true\|false\|null\)$END/\1/p" \
  -ne "s/\\\\\"/\\\\q/g;s/^.*$NAME\"\([^\"]*\)\"$END/\1/p")
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (json): \"$1\""
  log_notice_stack
  return $ERR_FATAL
 fi
 STR=${STR//\\\//\/}
 STR=${STR//\\\\/\\}
 STR=${STR//\\q/\"}
 STR=${STR//\\b/$'\b'}
 STR=${STR//\\f/$'\f'}
 STR=${STR//\\n/$'\n'}
 STR=${STR//\\r/$'\r'}
 STR=${STR//\\t/	}
 echo -e "$STR"
}
parse_json_quiet() {
 parse_json "$@" 2>/dev/null
 return 0
}
match_json_true() {
 grep -q "\"$1\"[[:space:]]*:[[:space:]]*true" <<< "$2"
}
grep_http_header_location() {
 parse_all '^[Ll]ocation:' 'n:[[:space:]]\+\(.*\)[[:cntrl:]]$'
}
grep_http_header_location_quiet() {
 parse_all '^[Ll]ocation:' 'n:[[:space:]]\+\(.*\)[[:cntrl:]]$' 2>/dev/null
 return 0
}
grep_http_header_content_location() {
 parse_all '^[Cc]ontent-[Ll]ocation:' 'n:[[:space:]]\+\(.*\)[[:cntrl:]]$'
}
grep_http_header_content_type() {
 parse_all '^[Cc]ontent-[Tt]ype:' 'e:[[:space:]]\+\(.*\)[[:cntrl:]]$'
}
grep_http_header_content_length() {
 parse_all '^[Cc]ontent-[Ll]ength:' 'h:[[:space:]]\+\(.*\)[[:cntrl:]]$'
}
grep_http_header_content_disposition() {
 parse_all '^[Cc]ontent-[Dd]isposition:' "filename\*\?=[\"']\?\([^\"'[:cntrl:]]*\)"
}
grep_form_by_name() {
 local -r A=${2:-'.*'}
 local STR=$(sed -ne \
  "/<[Ff][Oo][Rr][Mm][[:space:]].*name[[:space:]]*=[[:space:]]*[\"']\?$A[\"']\?[[:space:]/>]/,/<\/[Ff][Oo][Rr][Mm]>/p" <<< "$1")
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"name=$A\""
  return $ERR_FATAL
 fi
 echo "$STR"
}
grep_form_by_id() {
 local -r A=${2:-'.*'}
 local STR=$(sed -ne \
  "/<[Ff][Oo][Rr][Mm][[:space:]].*id[[:space:]]*=[[:space:]]*[\"']\?$A[\"']\?[[:space:]/>]/,/<\/[Ff][Oo][Rr][Mm]>/p" <<< "$1")
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"id=$A\""
  return $ERR_FATAL
 fi
 echo "$STR"
}
grep_form_by_order() {
 grep_block_by_order '[Ff][Oo][Rr][Mm]' "$@"
}
grep_script_by_order() {
 grep_block_by_order '[Ss][Cc][Rr][Ii][Pp][Tt]' "$@"
}
break_html_lines() {
 sed -e 's/<\/[^>]*>/&\n/g'
}
break_html_lines_alt() {
 sed -e 's/<[^>]*>/&\n/g'
}
strip_html_comments() {
 sed -e '/<!--/!b' -e ':a' -e '/-->/!{N;ba;}' -e 's/<!--.*-->//'
}
parse_all_tag() {
 local -r T=${2:-"$1"}
 local -r D=$'\001'
 local STR=$(sed -ne \
  "\\${D}$1${D}"'!b; s/<\/'"$T>.*//; s/^.*<$T\(>\|[[:space:]][^>]*>\)//p")
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"/$1/ <$T>\""
  log_notice_stack
  return $ERR_FATAL
 fi
 echo "$STR"
}
parse_all_tag_quiet() {
 parse_all_tag "$@" 2>/dev/null
 return 0
}
parse_tag() {
 local -r T=${2:-"$1"}
 local -r D=$'\001'
 local STR=$(sed -ne \
  "\\${D}$1${D}"'!b; s/<\/'"$T>.*//; s/^.*<$T\(>\|[[:space:]][^>]*>\)//;ta;b;:a;p;q;")
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"/$1/ <$T>\""
  log_notice_stack
  return $ERR_FATAL
 fi
 echo "$STR"
}
parse_tag_quiet() {
 parse_tag "$@" 2>/dev/null
 return 0
}
parse_all_attr() {
 local -r A=${2:-"$1"}
 local -r D=$'\001'
 local STR=$(sed \
  -ne "\\${D}$1${D}s${D}.*[[:space:]]\($A\)[[:space:]]*=[[:space:]]*\"\([^\">]*\).*${D}\2${D}p" \
  -ne "\\${D}$1${D}s${D}.*[[:space:]]\($A\)[[:space:]]*=[[:space:]]*'\([^'>]*\).*${D}\2${D}p" \
  -ne "\\${D}$1${D}s${D}.*[[:space:]]\($A\)[[:space:]]*=[[:space:]]*\([^[:space:]\"\`'<=>]\+\).*${D}\2${D}p")
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"/$1/ $A=\""
  log_notice_stack
  return $ERR_FATAL
 fi
 echo "$STR"
}
parse_all_attr_quiet() {
 parse_all_attr "$@" 2>/dev/null
 return 0
}
parse_attr() {
 local -r A=${2:-"$1"}
 local -r D=$'\001'
 local STR=$(sed \
  -ne "\\${D}$1${D}s${D}.*[[:space:]]\($A\)[[:space:]]*=[[:space:]]*\"\([^\">]*\).*${D}\2${D}p;ta" \
  -ne "\\${D}$1${D}s${D}.*[[:space:]]\($A\)[[:space:]]*=[[:space:]]*'\([^'>]*\).*${D}\2${D}p;ta" \
  -ne "\\${D}$1${D}s${D}.*[[:space:]]\($A\)[[:space:]]*=[[:space:]]*\([^[:space:]\"\`'<=>]\+\).*${D}\2${D}p;ta" \
  -ne 'b;:a;q;')
 if [ -z "$STR" ]; then
  log_error "$FUNCNAME failed (sed): \"/$1/ $A=\""
  log_notice_stack
  return $ERR_FATAL
 fi
 echo "$STR"
}
parse_attr_quiet() {
 parse_attr "$@" 2>/dev/null
 return 0
}
parse_form_action() {
 parse_attr '<[Ff][Oo][Rr][Mm]' 'action'
}
parse_form_input_by_name() {
 parse_attr "<[Ii][Nn][Pp][Uu][Tt][^>]*name[[:space:]]*=[[:space:]]*[\"']\?$1[\"']\?[[:space:]/>]" 'value'
}
parse_form_input_by_name_quiet() {
 parse_form_input_by_name "$@" 2>/dev/null
 return 0
}
parse_form_input_by_type() {
 parse_attr "<[Ii][Nn][Pp][Uu][Tt][^>]*type[[:space:]]*=[[:space:]]*[\"']\?$1[\"']\?[[:space:]/>]" 'value'
}
parse_form_input_by_type_quiet() {
 parse_form_input_by_type "$@" 2>/dev/null
 return 0
}
parse_form_input_by_id() {
 parse_attr "<[Ii][Nn][Pp][Uu][Tt][^>]*id[[:space:]]*=[[:space:]]*[\"']\?$1[\"']\?[[:space:]/>]" 'value'
}
parse_form_input_by_id_quiet() {
 parse_form_input_by_id "$@" 2>/dev/null
 return 0
}
parse_cookie() {
 parse_all "\t$1\t[^\t]*\$" "\t$1\t\(.*\)"
}
parse_cookie_quiet() {
 parse_all "\t$1\t[^\t]*\$" "\t$1\t\(.*\)" 2>/dev/null
 return 0
}
basename_url() {
 if [[ $1 =~ ([Hh][Tt][Tt][Pp][Ss]?|[Ff][Tt][Pp][Ss]?|[Ff][Ii][Ll][Ee])://[^/?#]* ]]; then
  echo "${BASH_REMATCH[0]}"
 else
  echo "$1"
 fi
}
basename_file() {
 basename -- "$1" || return $ERR_SYSTEM
}
html_to_utf8() {
 if check_exec 'recode'; then
  log_report "$FUNCNAME: use recode"
  recode html..utf8
 elif check_exec 'perl'; then
  log_report "$FUNCNAME: use perl"
  perl -n -mHTML::Entities \
   -e 'BEGIN { eval { binmode(STDOUT,q[:utf8]); }; } \
    print HTML::Entities::decode_entities($_);' 2>/dev/null || { \
   log_debug "$FUNCNAME failed (perl): HTML::Entities missing ?";
   cat;
  }
 else
  log_notice 'recode binary not found, pass-through'
  cat
 fi
}
uri_encode_strict() {
 sed -e '
s/\%/%25/g
s/ /%20/g
s/!/%21/g
s/#/%23/g
s/\$/%24/g
s/&/%26/g
s/'\''/%27/g
s/(/%28/g
s/)/%29/g
s/*/%2A/g
s/+/%2B/g
s/,/%2C/g
s|/|%2F|g
s/:/%3A/g
s/;/%3B/g
s/=/%3D/g
s/?/%3F/g
s/@/%40/g
s/\[/%5B/g
s/\]/%5D/g' #'# Help vim syntax highlighting
}
uri_encode() {
 sed -e '
s/ /%20/g
s/\[/%5B/g
s/\]/%5D/g'
}
uri_decode() {
 sed -e '
s/%20/ /g
s/%21/!/g
s/%26/\&/g
s/%28/(/g
s/%29/)/g
s/%2B/+/g
s/%2C/,/g
s|%2F|/|g
s/%3A/:/g
s/%3D/=/g
s/%40/@/g
s/%5B/\[/g
s/%5D/\]/g'
}
get_filesize() {
 local FILE_SIZE=$(ls -l "$1" 2>/dev/null | sed -e 's/[[:space:]]\+/ /g' | cut -d' ' -f5)
 if [ -z "$FILE_SIZE" ]; then
  log_error "$FUNCNAME: error accessing \`$1'"
  return $ERR_SYSTEM
 fi
 echo "$FILE_SIZE"
}
create_tempfile() {
 local -r SUFFIX=$1
 local FILE="$TMPDIR/$(basename_file "$0").$$.$RANDOM$SUFFIX"
 if touch "$FILE" && chmod 600 "$FILE"; then
  echo "$FILE"
  return 0
 fi
 return $ERR_SYSTEM
}
prompt_for_password() {
 local PASSWORD
 log_notice 'No password specified, enter it now (1 hour timeout)'
 IFS= read -s -r -t 3600 -p 'Enter password: ' PASSWORD
 stderr
 test -z "$PASSWORD" && return $ERR_LINK_PASSWORD_REQUIRED
 echo "$PASSWORD"
}
post_login() {
 local -r AUTH=$1
 local -r COOKIE=$2
 local -r POSTDATA=$3
 local -r LOGIN_URL=$4
 shift 4
 local -a CURL_ARGS=("$@")
 local USER PASSWORD DATA RESULT
 if [ -z "$AUTH" ]; then
  log_error "$FUNCNAME: authentication string is empty"
  return $ERR_LOGIN_FAILED
 fi
 if [ -z "$COOKIE" ]; then
  log_error "$FUNCNAME: cookie file expected"
  return $ERR_LOGIN_FAILED
 fi
 USER=$(echo "${AUTH%%:*}" | uri_encode_strict)
 PASSWORD=$(echo "${AUTH#*:}" | uri_encode_strict)
 if [ -z "$PASSWORD" -o "$AUTH" = "${AUTH#*:}" ]; then
  PASSWORD=$(prompt_for_password) || true
 fi
 log_notice "Starting login process: $USER/${PASSWORD//?/*}"
 DATA=$(eval echo "${POSTDATA//&/\\&}")
 RESULT=$(curl --cookie-jar "$COOKIE" --data "$DATA" "${CURL_ARGS[@]}" \
  "$LOGIN_URL") || return
 if [ ! -s "$COOKIE" ]; then
  log_debug "$FUNCNAME: no entry was set (empty cookie file)"
  return $ERR_LOGIN_FAILED
 fi
 log_report '=== COOKIE BEGIN ==='
 logcat_report "$COOKIE"
 log_report '=== COOKIE END ==='
 if ! find_in_array CURL_ARGS[@] '-o' '--output'; then
  echo "$RESULT"
 fi
}
debian_javascript_enabled() {
 if [ "$PLOWCORE_DEBIAN_JS" != 'yes' ]; then
  return $ERR_SYSTEM
 fi
}
detect_javascript() {
 if ! debian_javascript_enabled; then
  log_notice "Plowshare's use of Javascript interpreter is disabled in Debian by default for security. (See: #791467)"
  log_notice 'Javascript may be enabled by setting the environment variable PLOWSHARE_DEBIAN_JS=yes'
  return $ERR_SYSTEM
 fi
 if ! type -P "$PLOWCORE_JS" >/dev/null 2>&1; then
  log_notice 'Javascript interpreter not found. Please install one!'
  return $ERR_SYSTEM
 fi
}
javascript() {
 local TEMPSCRIPT
 detect_javascript || return
 TEMPSCRIPT=$(create_tempfile '.js') || return
 cat > "$TEMPSCRIPT"
 log_report "interpreter: $(type -P js)"
 log_report '=== JAVASCRIPT BEGIN ==='
 logcat_report "$TEMPSCRIPT"
 log_report '=== JAVASCRIPT END ==='
 command "$PLOWCORE_JS" "$TEMPSCRIPT"
 rm -f "$TEMPSCRIPT"
 return 0
}
wait() {
 local -r VALUE=$1
 local -r UNIT=$2
 local UNIT_STR TOTAL_SECS
 if test "$VALUE" = '0'; then
  log_debug 'wait called with null duration'
  return 0
 fi
 if [ "$UNIT" = 'minutes' ]; then
  UNIT_STR=minutes
  TOTAL_SECS=$((VALUE * 60))
 else
  UNIT_STR=seconds
  TOTAL_SECS=$((VALUE))
 fi
 timeout_update $TOTAL_SECS || return
 local REMAINING=$TOTAL_SECS
 local MSG="Waiting $VALUE $UNIT_STR..."
 local CLEAR="     \b\b\b\b\b"
 if test -t 2; then
  local START_DATE=$(date +%s)
  CONT_SIGNAL=
  while [ "$REMAINING" -gt 0 ]; do
   log_notice_norc "\r$MSG $(splitseconds $REMAINING) left$CLEAR"
   sleep 1
   if [[ $CONT_SIGNAL ]]; then
    local -i TMP_SECS
    (( TMP_SECS = TOTAL_SECS - (CONT_SIGNAL - START_DATE) ))
    (( TMP_SECS >= 0 )) || TMP_SECS=0
    CONT_SIGNAL=
    log_debug "SIGCONT detected, adjust wait time ($REMAINING => $TMP_SECS)"
    REMAINING=$TMP_SECS
   else
    (( --REMAINING ))
   fi
  done
  log_notice_norc "\r$MSG done$CLEAR\n"
 else
  log_notice "$MSG"
  sleep $TOTAL_SECS
 fi
}
captcha_process() {
 local -r CAPTCHA_TYPE=$2
 local METHOD_SOLVE METHOD_VIEW FILENAME RESPONSE WORD I FBDEV
 local TID=0
 if [ -f "$1" ]; then
  FILENAME=$1
 elif match_remote_url "$1"; then
  FILENAME=$(create_tempfile '.captcha') || return
  curl -o "$FILENAME" "$1" || return
 else
  log_error "$FUNCNAME: image file not found"
  return $ERR_FATAL
 fi
 if [ ! -s "$FILENAME" ]; then
  log_error "$FUNCNAME: empty image file"
  return $ERR_FATAL
 fi
 if [ -n "$CAPTCHA_PROGRAM" ]; then
  local RET=0
  WORD=$(exec "$CAPTCHA_PROGRAM" "$MODULE" "$FILENAME" "${CAPTCHA_TYPE}${3:+-$3}") || RET=$?
  if [ $RET -eq 0 ]; then
   echo "$WORD"
   echo $TID
   return 0
  elif [ $RET -ne $ERR_NOMODULE ]; then
   log_error "captchaprogram exit with status $RET"
   return $RET
  fi
 fi
 if [ -n "$CAPTCHA_METHOD" ]; then
  captcha_method_translate "$CAPTCHA_METHOD" METHOD_SOLVE METHOD_VIEW
 else
  METHOD_SOLVE='online,prompt'
 fi
 if [[ $METHOD_SOLVE = *online* ]]; then
  if service_antigate_ready "$CAPTCHA_ANTIGATE"; then
   METHOD_SOLVE=antigate
   : ${METHOD_VIEW:=log}
  elif service_9kweu_ready "$CAPTCHA_9KWEU"; then
   METHOD_SOLVE=9kweu
   : ${METHOD_VIEW:=log}
  elif service_captchabrotherhood_ready "${CAPTCHA_BHOOD%%:*}" "${CAPTCHA_BHOOD#*:}"; then
   METHOD_SOLVE=captchabrotherhood
   : ${METHOD_VIEW:=log}
  elif service_captchacoin_ready "$CAPTCHA_COIN"; then
   METHOD_SOLVE=captchacoin
   : ${METHOD_VIEW:=log}
  elif service_captchadeathby_ready "${CAPTCHA_DEATHBY%%:*}" "${CAPTCHA_DEATHBY#*:}"; then
   METHOD_SOLVE=deathbycaptcha
   : ${METHOD_VIEW:=log}
  elif [ -n "$CAPTCHA_METHOD" ]; then
   log_error 'No online recognition service found! Captcha solving request will fail.'
   METHOD_SOLVE=none
  fi
 fi
 if [ "$METHOD_SOLVE" != "${METHOD_SOLVE//,}" ]; then
  METHOD_SOLVE=${METHOD_SOLVE##*,}
  log_debug "no captcha solving method found: fallback to \`$METHOD_SOLVE'."
 fi
 if [[ "$(tty)" = /dev/tty* ]]; then
  : ${METHOD_VIEW:=view-x,view-fb,view-aa,log}
 else
  : ${METHOD_VIEW:=view-x,view-aa,log}
 fi
 if [[ $METHOD_VIEW = *view-x* ]]; then
  if test -z "$DISPLAY"; then
   log_notice 'DISPLAY variable not exported! Skip X11 viewers probing.'
  elif check_exec 'display'; then
   METHOD_VIEW=X-display
  elif check_exec 'feh'; then
   METHOD_VIEW=X-feh
  elif check_exec 'sxiv'; then
   METHOD_VIEW=X-sxiv
  elif check_exec 'qiv'; then
   METHOD_VIEW=X-qiv
  else
   log_notice 'No X11 image viewer found to display captcha image'
  fi
 fi
 if [[ $METHOD_VIEW = *view-fb* ]]; then
  if test -n "$FRAMEBUFFER"; then
   log_notice 'FRAMEBUFFER variable is not empty, use it.'
   FBDEV=$FRAMEBUFFER
  else
   FBDEV=/dev/fb0
  fi
  if ! test -c "$FBDEV"; then
   log_notice "$FBDEV not found! Skip FB viewers probing."
  elif check_exec 'fbi'; then
   METHOD_VIEW=fb-fbi
  elif check_exec 'fim'; then
   METHOD_VIEW=fb-fim
  else
   log_notice 'No FB image viewer found to display captcha image'
  fi
 fi
 local MAX_OUTPUT_WIDTH MAX_OUTPUT_HEIGHT
 if [[ $METHOD_VIEW = *view-aa* ]]; then
  if check_exec img2txt; then
   METHOD_VIEW=img2txt
  elif check_exec tiv; then
   METHOD_VIEW=tiv
  elif check_exec aview && check_exec convert; then
   METHOD_VIEW=aview
  else
   log_notice 'No ascii viewer found to display captcha image'
  fi
  if [[ $METHOD_VIEW != *view-aa* ]]; then
   if check_exec tput; then
    local TYPE
    if [ -z "$TERM" -o "$TERM" = 'dumb' ]; then
     log_notice 'Invalid $TERM value. Terminal type forced to vt100.'
     TYPE='-Tvt100'
    fi
    MAX_OUTPUT_WIDTH=$(tput $TYPE cols)
    MAX_OUTPUT_HEIGHT=$(tput $TYPE lines)
   else
    MAX_OUTPUT_WIDTH=${COLUMNS:-150}
    MAX_OUTPUT_HEIGHT=${LINES:-57}
   fi
   if check_exec identify; then
    local -i W H
    read -r W H < <(identify -quiet -format '%w %h' "$FILENAME")
    [ "$W" -lt "$MAX_OUTPUT_WIDTH" ] && MAX_OUTPUT_WIDTH=$W
    [ "$H" -lt "$MAX_OUTPUT_HEIGHT" ] && MAX_OUTPUT_HEIGHT=$H
   fi
  fi
 fi
 if [ "$METHOD_VIEW" != "${METHOD_VIEW//,}" ]; then
  METHOD_VIEW=${METHOD_VIEW##*,}
  log_debug "no captcha viewing method found: fallback to \`$METHOD_VIEW'."
 fi
 local IMG_HASH PRG_PID IMG_PNM
 case $METHOD_VIEW in
  log)
   log_notice "Local image: $FILENAME"
   ;;
  aview)
   local -r FF=$'\f'
   IMG_PNM=$(create_tempfile '.pnm') || return
   convert "$FILENAME" -negate -depth 8 pnm:$IMG_PNM && \
    aview -width $MAX_OUTPUT_WIDTH -height $MAX_OUTPUT_HEIGHT \
     -kbddriver stdin -driver stdout "$IMG_PNM" 2>/dev/null <<<'q' | \
      sed -e "1d;/$FF/,/$FF/d;/^[[:space:]]*$/d" 1>&2
   rm -f "$IMG_PNM"
   ;;
  tiv)
   tiv -a -w $MAX_OUTPUT_WIDTH -h $MAX_OUTPUT_HEIGHT "$FILENAME" 1>&2
   ;;
  img2txt)
   img2txt -W $MAX_OUTPUT_WIDTH -H $MAX_OUTPUT_HEIGHT "$FILENAME" 1>&2
   ;;
  X-display)
   display "$FILENAME" &
   PRG_PID=$!
   ;;
  X-feh)
   feh "$FILENAME" &
   PRG_PID=$!
   ;;
  X-qiv)
   qiv "$FILENAME" &
   PRG_PID=$!
   ;;
  X-sxiv)
   sxiv -q -s f "$FILENAME" &
   [ $? -eq 0 ] && PRG_PID=$!
   ;;
  fb-fbi)
   log_debug "fbi -autozoom -noverbose -d $FBDEV $FILENAME"
   fbi -autozoom -noverbose -d "$FBDEV" "$FILENAME" &>/dev/null
   PRG_PID=""
   ;;
  fb-fim)
   log_debug "fim --quiet --autozoom -d $FBDEV $FILENAME"
   fim --quiet --autozoom -d "$FBDEV" "$FILENAME" 2>/dev/null
   PRG_PID=""
   ;;
  imgur)
   IMG_HASH=$(image_upload_imgur "$FILENAME") || true
   ;;
  *)
   log_error "unknown view method: $METHOD_VIEW"
   rm -f "$FILENAME"
   return $ERR_FATAL
   ;;
 esac
 case $METHOD_SOLVE in
  none)
   rm -f "$FILENAME"
   return $ERR_CAPTCHA
   ;;
  9kweu)
   log_notice 'Using 9kw.eu captcha recognition system'
   RESPONSE=$(curl -F 'method=post' \
    -F 'action=usercaptchaupload' \
    -F "apikey=$CAPTCHA_9KWEU" \
    -F 'source=plowshare' \
    -F 'maxtimeout=235' \
    -F "file-upload-01=@$FILENAME;filename=file.jpg" \
    'http://www.9kw.eu/index.cgi') || return
   if [ -z "$RESPONSE" ]; then
    log_error '9kw.eu empty answer'
    rm -f "$FILENAME"
    return $ERR_NETWORK
   elif [[ $RESPONSE = 00[012][[:digit:]][[:space:]]* ]]; then
    log_error "9kw.eu error: ${RESPONSE:5}"
    rm -f "$FILENAME"
    return $ERR_FATAL
   fi
   TID=$RESPONSE
   for I in 10 6 6 7 7 8 8 9 9 10 10 20 20 20 30 30 30; do
    wait $I seconds
    RESPONSE=$(curl --get --data 'action=usercaptchacorrectdata' \
     --data "apikey=$CAPTCHA_9KWEU" \
     --data "id=$TID" \
     --data 'source=plowshare' --data 'info=1' \
     'http://www.9kw.eu/index.cgi') || return
    if [ 'NO DATA' = "$RESPONSE" ]; then
     continue
    elif [ -z "$RESPONSE" ]; then
     continue
    elif [ -n "$RESPONSE" ]; then
     WORD=$RESPONSE
     break
    else
     log_error "9kw.eu error: $RESPONSE"
     rm -f "$FILENAME"
     return $ERR_FATAL
    fi
   done
   if [ -z "$WORD" ]; then
    log_error '9kw.eu error: service not available'
    rm -f "$FILENAME"
    return $ERR_CAPTCHA
   fi
   echo "$WORD"
   echo "9$TID"
   ;;
  antigate)
   log_notice 'Using antigate captcha recognition system'
   RESPONSE=$(curl -F 'method=post' \
    -F "file=@$FILENAME;filename=file.jpg" \
    -F "key=$CAPTCHA_ANTIGATE" \
    -F 'is_russian=0' \
    'http://antigate.com/in.php') || return
   if [ -z "$RESPONSE" ]; then
    log_error 'antigate empty answer'
    rm -f "$FILENAME"
    return $ERR_NETWORK
   elif [ 'ERROR_IP_NOT_ALLOWED' = "$RESPONSE" ]; then
    log_error 'antigate error: IP not allowed'
    rm -f "$FILENAME"
    return $ERR_FATAL
   elif [ 'ERROR_ZERO_BALANCE' = "$RESPONSE" ]; then
    log_error 'antigate error: no credits'
    rm -f "$FILENAME"
    return $ERR_FATAL
   elif [ 'ERROR_NO_SLOT_AVAILABLE' = "$RESPONSE" ]; then
    log_error 'antigate error: no slot available'
    rm -f "$FILENAME"
    return $ERR_CAPTCHA
   elif match 'ERROR_' "$RESPONSE"; then
    log_error "antigate error: $RESPONSE"
    rm -f "$FILENAME"
    return $ERR_FATAL
   fi
   TID=$(echo "$RESPONSE" | parse_quiet . 'OK|\(.*\)')
   for I in 8 5 5 6 6 7 7 8; do
    wait $I seconds
    RESPONSE=$(curl --get \
     --data "key=${CAPTCHA_ANTIGATE}&action=get&id=$TID" \
     'http://antigate.com/res.php') || return
    if [ 'CAPCHA_NOT_READY' = "$RESPONSE" ]; then
     continue
    elif match '^OK|' "$RESPONSE"; then
     WORD=$(echo "$RESPONSE" | parse_quiet . 'OK|\(.*\)')
     break
    else
     log_error "antigate error: $RESPONSE"
     rm -f "$FILENAME"
     return $ERR_FATAL
    fi
   done
   if [ -z "$WORD" ]; then
    log_error 'antigate error: service not available'
    rm -f "$FILENAME"
    return $ERR_CAPTCHA
   fi
   echo "$WORD"
   echo "a$TID"
   ;;
  captchabrotherhood)
   local USERNAME=${CAPTCHA_BHOOD%%:*}
   local PASSWORD=${CAPTCHA_BHOOD#*:}
   log_notice "Using captcha brotherhood bypass service ($USERNAME)"
   RESPONSE=$(curl --data-binary "@$FILENAME" \
    --header 'Content-Type: text/html'     \
    "http://www.captchabrotherhood.com/sendNewCaptcha.aspx?username=$USERNAME&password=$PASSWORD&captchaSource=plowshare&timeout=30&captchaSite=-1") || return
   if [ "${RESPONSE:0:3}" = 'OK-' ]; then
    TID=${RESPONSE:3}
    if [ -n "$TID" ]; then
     for I in 6 5 5 6 6 7 7 8; do
      wait $I seconds
      RESPONSE=$(curl --get -d "username=$USERNAME"   \
       -d "password=$PASSWORD" -d "captchaID=$TID" \
       'http://www.captchabrotherhood.com/askCaptchaResult.aspx') || return
      if [ "${RESPONSE:0:12}" = 'OK-answered-' ]; then
       WORD=${RESPONSE:12}
       if [ -n "$WORD" ]; then
        echo "$WORD"
        echo "b$TID"
        return 0
       else
        RESPONSE='empty word?'
       fi
       break
      elif [ "${RESPONSE:0:3}" != 'OK-' ]; then
       break
      fi
     done
    else
     RESPONSE='empty tid?'
    fi
   fi
   log_error "Captcha Brotherhood error: ${RESPONSE#Error-}"
   rm -f "$FILENAME"
   return $ERR_FATAL
   ;;
  captchacoin)
   log_notice 'Using CaptchaCoin captcha recognition system'
   RESPONSE=$(curl -F "file=@$FILENAME;filename=file.jpg" \
    -F "api_key=$CAPTCHA_COIN" \
    'http://www.captchacoin.com/api/submit') || return
   if [ -z "$RESPONSE" ]; then
    log_error 'CaptchaCoin empty answer'
    rm -f "$FILENAME"
    return $ERR_NETWORK
   fi
   if match_json_true 'success' "$RESPONSE"; then
    TID=$(echo "$RESPONSE" | parse_json_quiet 'img_key')
    if [ -z "$TID" ]; then
     log_error 'CaptchaCoin error: no image key'
     rm -f "$FILENAME"
     return $ERR_FATAL
    fi
   else
    log_error "CaptchaCoin error: $RESPONSE"
    rm -f "$FILENAME"
    return $ERR_FATAL
   fi
   for I in 10 5 5 6 6 7 7 8; do
    wait $I seconds
    RESPONSE=$(curl --get \
     -d "api_key=$CAPTCHA_COIN" \
     -d "img_key=$TID" \
     'http://api.captchacoin.com/api/imginfo') || return
    if match_json_true 'success' "$RESPONSE"; then
     if match_json_true 'assigned' "$RESPONSE"; then
      log_debug 'CaptchaCoin: someone is solving the captcha'
     fi
     CAPCTHA_AGE=$(echo "$RESPONSE" | parse_json_quiet 'age_seconds')
     [ -n "$CAPCTHA_AGE" ] && log_debug "CaptchaCoin: captcha age $CAPCTHA_AGE"
     WORD=$(echo "$RESPONSE" | parse_json_quiet 'solution')
     [ -n "$WORD" ] && break
    else
     log_error "CaptchaCoin error: $RESPONSE"
     rm -f "$FILENAME"
     return $ERR_FATAL
    fi
   done
   if [ -z "$WORD" ]; then
    RESPONSE=$(curl --get \
     -d "api_key=$CAPTCHA_COIN" \
     -d "img_key=$TID" \
     'http://www.captchacoin.com/api/withdraw') || return
    if match_json_true 'success' "$RESPONSE"; then
     log_error 'CaptchaCoin: recognition failed, withdraw successful'
    else
     log_error "CaptchaCoin: recognition failed, withdraw error ($RESPONSE)"
    fi
    rm -f "$FILENAME"
    return $ERR_FATAL
   fi
   echo "$WORD"
   echo "c$TID"
   ;;
  deathbycaptcha)
   local HTTP_CODE POLL_URL
   local USERNAME=${CAPTCHA_DEATHBY%%:*}
   local PASSWORD=${CAPTCHA_DEATHBY#*:}
   log_notice "Using DeathByCaptcha service ($USERNAME)"
   RESPONSE=$(curl --include --header 'Expect: ' \
    --header 'Accept: application/json' \
    -F "username=$USERNAME" \
    -F "password=$PASSWORD" \
    -F "captchafile=@$FILENAME" \
    'http://api.dbcapi.me/api/captcha') || return
   if [ -z "$RESPONSE" ]; then
    log_error 'DeathByCaptcha empty answer'
    rm -f "$FILENAME"
    return $ERR_NETWORK
   fi
   HTTP_CODE=$(echo "$RESPONSE" | first_line | \
    parse . 'HTTP/1\.. \([[:digit:]]\+\) ')
   if [ "$HTTP_CODE" = 303 ]; then
    POLL_URL=$(echo "$RESPONSE" | grep_http_header_location) || return
    for I in 4 3 3 4 4 5 5; do
     wait $I seconds
     RESPONSE=$(curl --header 'Accept: application/json' \
      "$POLL_URL") || return
     if match_json_true 'is_correct' "$RESPONSE"; then
      WORD=$(echo "$RESPONSE" | parse_json_quiet text)
      if [ -n "$WORD" ]; then
       TID=$(echo "$RESPONSE" | parse_json_quiet captcha)
       echo "$WORD"
       echo "d$TID"
       return 0
      fi
     else
      log_error "DeathByCaptcha unknown error: $RESPONSE"
      rm -f "$FILENAME"
      return $ERR_CAPTCHA
     fi
    done
    log_error 'DeathByCaptcha timeout: give up!'
   else
    log_error "DeathByCaptcha wrong http answer ($HTTP_CODE)"
   fi
   rm -f "$FILENAME"
   return $ERR_CAPTCHA
   ;;
  prompt)
   if [ "$CAPTCHA_TYPE" = 'recaptcha' -o \
     "$CAPTCHA_TYPE" = 'solvemedia' ]; then
    log_notice 'Leave this field blank and hit enter to get another captcha image'
   fi
   read -r -p 'Enter captcha response (drop punctuation marks, case insensitive): ' RESPONSE
   echo "$RESPONSE"
   echo $TID
   ;;
  *)
   log_error "unknown solve method: $METHOD_SOLVE"
   rm -f "$FILENAME"
   return $ERR_FATAL
   ;;
 esac
 case $METHOD_VIEW in
  X-*)
   [[ $PRG_PID ]] && kill -HUP $PRG_PID >/dev/null 2>&1
   ;;
  imgur)
   image_delete_imgur "$IMG_HASH" || true
   ;;
 esac
 if [ "$1" != "$FILENAME" ]; then
  rm -f "$FILENAME"
 fi
}
recaptcha_process() {
 local -r RECAPTCHA_SERVER='http://www.google.com/recaptcha/api/'
 local URL="${RECAPTCHA_SERVER}challenge?k=${1}"
 local VARS SERVER TRY CHALLENGE FILENAME WORDS TID
 VARS=$(curl -L "$URL") || return
 if [ -z "$VARS" ]; then
  return $ERR_CAPTCHA
 fi
 SERVER=$(echo "$VARS" | parse_quiet 'server' "server[[:space:]]\?:[[:space:]]\?'\([^']*\)'") || return
 CHALLENGE=$(echo "$VARS" | parse_quiet 'challenge' "challenge[[:space:]]\?:[[:space:]]\?'\([^']*\)'") || return
 VARS=$(curl "${SERVER}reload?k=${1}&c=${CHALLENGE}&reason=i&type=image&lang=en") || return
 CHALLENGE=$(echo "$VARS" | parse 'finish_reload' "('\([^']*\)") || return
 log_debug "reCaptcha server: $SERVER"
 FILENAME=$(create_tempfile '.recaptcha.jpg') || return
 TRY=0
 while (( TRY++ < 100 )) || return $ERR_MAX_TRIES_REACHED; do
  log_debug "reCaptcha loop $TRY"
  log_debug "reCaptcha challenge: $CHALLENGE"
  URL="${SERVER}image?c=${CHALLENGE}"
  log_debug "reCaptcha image URL: $URL"
  curl "$URL" -o "$FILENAME" || return
  WORDS=$(captcha_process "$FILENAME" recaptcha) || return
  rm -f "$FILENAME"
  { read WORDS; read TID; } <<< "$WORDS"
  [ -n "$WORDS" ] && break
  log_debug 'empty, request another image'
  VARS=$(curl "${SERVER}reload?k=${1}&c=${CHALLENGE}&reason=r&type=image&lang=en") || return
  CHALLENGE=$(echo "$VARS" | parse 'finish_reload' "('\([^']*\)") || return
 done
 WORDS=$(echo "$WORDS" | uri_encode)
 echo "$WORDS"
 echo "$CHALLENGE"
 echo $TID
}
solvemedia_captcha_process() {
 local -r PUB_KEY=$1
 local -r BASE_URL='http://api.solvemedia.com/papi'
 local URL="$BASE_URL/challenge.noscript?k=$PUB_KEY"
 local HTML MAGIC CHALL IMG_FILE XY WI WORDS TID TRY
 IMG_FILE=$(create_tempfile '.solvemedia.jpg') || return
 TRY=0
 while (( TRY++ < 100 )) || return $ERR_MAX_TRIES_REACHED; do
  log_debug "SolveMedia loop $TRY"
  XY=''
  HTML=$(curl "$URL") || return
  MAGIC=$(echo "$HTML" | parse_form_input_by_name 'magic') || return
  CHALL=$(echo "$HTML" | parse_form_input_by_name \
   'adcopy_challenge') || return
  curl -o "$IMG_FILE" "$BASE_URL/media?c=$CHALL" || return
  WI=$(captcha_process "$IMG_FILE" solvemedia) || return
  { read WORDS; read TID; } <<< "$WI"
  rm -f "$IMG_FILE"
  if [ -z "$WORDS" ]; then
   log_debug 'empty, request another image'
   XY='-d t_img.x=23 -d t_img.y=7'
  fi
  HTML=$(curl --referer "$URL" \
   -d "adcopy_response=$WORDS" \
   -d "k=$PUB_KEY" \
   -d 'l=en' \
   -d 't=img' \
   -d 's=standard' \
   -d "magic=$MAGIC" \
   -d "adcopy_challenge=$CHALL" \
   $XY \
   "$BASE_URL/verify.noscript") || return
  if ! match 'Redirecting\.\.\.' "$HTML" ||
   match '&error=1&' "$HTML"; then
   captcha_nack "$TID"
   return $ERR_CAPTCHA
  fi
  URL=$(echo "$HTML" | parse 'META' 'URL=\(.\+\)">') || return
  [ -n "$WORDS" ] && break
 done
 HTML=$(curl "$URL") || return
 if ! match 'Please copy this gibberish:' "$HTML" || \
   ! match "$CHALL" "$HTML"; then
  log_debug 'Unexpected content. Site updated?'
  return $ERR_FATAL
 fi
 echo "$CHALL"
 echo "$TID"
}
captcha_ack() {
 [ "$1" = 0 ] && return
 local -r M=${1:0:1}
 local -r TID=${1:1}
 local RESPONSE STR
 if [ '9' = "$M" ]; then
  if [ -n "$CAPTCHA_9KWEU" ]; then
   RESPONSE=$(curl --get --data 'action=usercaptchacorrectback' \
    --data "apikey=$CAPTCHA_9KWEU" --data "id=$TID" \
    --data 'correct=1' 'http://www.9kw.eu/index.cgi') || return
   [ 'OK' = "$RESPONSE" ] || \
    log_error "9kw.eu error: $RESPONSE"
  else
   log_error "$FUNCNAME failed: 9kweu missing captcha key"
  fi
 elif [ c = "$M" ]; then
  if [ -n "$CAPTCHA_COIN" ]; then
   log_debug 'CaptchaCoin report ack'
   RESPONSE=$(curl --get \
    -d "api_key=$CAPTCHA_COIN" -d 'status=1' \
    -d "img_key=$TID" \
    'http://www.captchacoin.com/api/poststatus') || return
   if ! match_json_true 'success' "$RESPONSE"; then
    log_error "CaptchaCoin: report ack error ($RESPONSE)"
   fi
  else
   log_error "$FUNCNAME failed: CaptchaCoin missing API key"
  fi
 elif [[ $M != [abd] ]]; then
  log_error "$FUNCNAME failed: unknown transaction ID: $1"
 fi
}
captcha_nack() {
 [ "$1" = 0 ] && return
 local -r M=${1:0:1}
 local -r TID=${1:1}
 local RESPONSE STR
 if [ '9' = "$M" ]; then
  if [ -n "$CAPTCHA_9KWEU" ]; then
   RESPONSE=$(curl --get --data 'action=usercaptchacorrectback' \
    --data "apikey=$CAPTCHA_9KWEU" --data "id=$TID" \
    --data 'correct=2' 'http://www.9kw.eu/index.cgi') || return
   [ 'OK' = "$RESPONSE" ] || \
    log_error "9kw.eu error: $RESPONSE"
  else
   log_error "$FUNCNAME failed: 9kweu missing captcha key"
  fi
 elif [ a = "$M" ]; then
  if [ -n "$CAPTCHA_ANTIGATE" ]; then
   RESPONSE=$(curl --get \
    --data "key=${CAPTCHA_ANTIGATE}&action=reportbad&id=$TID"  \
    'http://antigate.com/res.php') || return
   [ 'OK_REPORT_RECORDED' = "$RESPONSE" ] || \
    log_error "antigate error: $RESPONSE"
  else
   log_error "$FUNCNAME failed: antigate missing captcha key"
  fi
 elif [ b = "$M" ]; then
  if [ -n "$CAPTCHA_BHOOD" ]; then
   local USERNAME=${CAPTCHA_BHOOD%%:*}
   local PASSWORD=${CAPTCHA_BHOOD#*:}
   log_debug "captcha brotherhood report nack ($USERNAME)"
   RESPONSE=$(curl --get \
    -d "username=$USERNAME" -d "password=$PASSWORD" \
    -d "captchaID=$TID" \
    'http://www.captchabrotherhood.com/complainCaptcha.aspx') || return
   [ 'OK-Complained' = "$RESPONSE" ] || \
    log_error "$FUNCNAME FIXME cbh[$RESPONSE]"
  else
   log_error "$FUNCNAME failed: captcha brotherhood missing account data"
  fi
 elif [ c = "$M" ]; then
  if [ -n "$CAPTCHA_COIN" ]; then
   log_debug 'CaptchaCoin report nack'
   RESPONSE=$(curl --get \
    -d "api_key=$CAPTCHA_COIN" -d 'status=0' \
    -d "img_key=$TID" \
    'http://www.captchacoin.com/api/poststatus') || return
   if ! match_json_true 'success' "$RESPONSE"; then
    log_error "CaptchaCoin: report nack error ($RESPONSE)"
   fi
  else
   log_error "$FUNCNAME failed: CaptchaCoin missing API key"
  fi
 elif [ d = "$M" ]; then
  if [ -n "$CAPTCHA_DEATHBY" ]; then
   local USERNAME=${CAPTCHA_DEATHBY%%:*}
   local PASSWORD=${CAPTCHA_DEATHBY#*:}
   log_debug "DeathByCaptcha report nack ($USERNAME)"
   RESPONSE=$(curl -F "username=$USERNAME" -F "password=$PASSWORD" \
    --header 'Accept: application/json' \
    "http://api.dbcapi.me/api/captcha/$TID/report") || return
   STR=$(echo "$RESPONSE" | parse_json_quiet 'status')
   [ "$STATUS" = '0' ] || \
    log_error "DeathByCaptcha: report nack error ($RESPONSE)"
  else
   log_error "$FUNCNAME failed: DeathByCaptcha missing account data"
  fi
 else
  log_error "$FUNCNAME failed: unknown transaction ID: $1"
 fi
}
random() {
 local -i I=0
 local LEN=${2:-8}
 local -r SEED=$RANDOM
 local RESULT N L
 case $1 in
  d|dec)
   RESULT=$(( SEED % 9 + 1 ))
   (( ++I ))
   while (( I < $LEN )); do
    printf -v N '%04u' $((RANDOM % 10000))
    RESULT+=$N
    (( I += 4 ))
   done
   ;;
  h|hex)
   printf -v RESULT '%x' $(( SEED % 15 + 1 ))
   (( ++I ))
   while (( I < $LEN )); do
    printf -v N '%04x' $((RANDOM & 65535))
    RESULT+=$N
    (( I += 4 ))
   done
   ;;
  H|HEX)
   printf -v RESULT '%X' $(( SEED % 15 + 1 ))
   (( ++I ))
   while (( I < $LEN )); do
    printf -v N '%04X' $((RANDOM & 65535))
    RESULT+=$N
    (( I += 4 ))
   done
   ;;
  l)
   while (( I++ < $LEN )); do
    N=$(( RANDOM % 26 + 16#61))
    printf -v L \\$(($N/64*100+$N%64/8*10+$N%8))
    RESULT+=$L
   done
   ;;
  L)
   while (( I++ < $LEN )); do
    N=$(( RANDOM % 26 + 16#41))
    printf -v L \\$(($N/64*100+$N%64/8*10+$N%8))
    RESULT+=$L
   done
   ;;
  [Ll][Ll])
   while (( I++ < $LEN )); do
    N=$(( RANDOM % 52 + 16#41))
    [[ $N -gt 90 ]] && (( N += 6 ))
    printf -v L \\$(($N/64*100+$N%64/8*10+$N%8))
    RESULT+=$L
   done
   ;;
  a)
   while (( I++ < $LEN )); do
    N=$(( RANDOM % 36 + 16#30))
    [[ $N -gt 57 ]] && (( N += 39 ))
    printf -v L \\$(($N/64*100+$N%64/8*10+$N%8))
    RESULT+=$L
   done
   ;;
  js)
   LEN=$((SEED % 3 + 17))
   RESULT='0.'$((RANDOM * 69069 & 16#ffffffff))
   RESULT+=$((RANDOM * 69069 & 16#ffffffff))
   ;;
  u16)
   RESULT=$(( 256 * (SEED & 255) + (RANDOM & 255) ))
   LEN=${#RESULT}
   ;;
  *)
   log_error "$FUNCNAME: unknown operation '$1'"
   return $ERR_FATAL
   ;;
 esac
 echo ${RESULT:0:$LEN}
}
md5() {
 if check_exec md5sum; then
  echo -n "$1" | md5sum -b 2>/dev/null | cut -d' ' -f1
 elif check_exec md5; then
  command md5 -qs "$1"
 elif check_exec openssl; then
  echo -n "$1" | openssl dgst -md5 | cut -d' ' -f2
 else
  log_error "$FUNCNAME: cannot find md5 calculator"
  return $ERR_SYSTEM
 fi
}
sha1() {
 if check_exec sha1sum; then
  echo -n "$1" | sha1sum -b 2>/dev/null | cut -d' ' -f1
 elif check_exec sha1; then
  command sha1 -qs "$1"
 elif check_exec openssl; then
  echo -n "$1" | openssl dgst -sha1 | cut -d' ' -f2
 else
  log_error "$FUNCNAME: cannot find sha1 calculator"
  return $ERR_SYSTEM
 fi
}
md5_file() {
 if [ -f "$1" ]; then
  if check_exec md5sum; then
   md5sum -b "$1" 2>/dev/null | cut -d' ' -f1
  elif check_exec md5; then
   command md5 -q "$1"
  elif check_exec openssl; then
   openssl dgst -md5 "$1" | cut -d' ' -f2
  else
   log_error "$FUNCNAME: cannot find md5 calculator"
   return $ERR_SYSTEM
  fi
 else
  log_error "$FUNCNAME: cannot stat file"
  return $ERR_SYSTEM
 fi
}
sha1_file() {
 if [ -f "$1" ]; then
  if check_exec sha1sum; then
   sha1sum -b "$1" 2>/dev/null | cut -d' ' -f1
  elif check_exec sha1; then
   command sha1 -q "$1"
  elif check_exec openssl; then
   openssl dgst -sha1 "$1" | cut -d' ' -f2
  else
   log_error "$FUNCNAME: cannot find sha1 calculator"
   return $ERR_SYSTEM
  fi
 else
  log_error "$FUNCNAME: cannot stat file"
  return $ERR_SYSTEM
 fi
}
split_auth() {
 local __AUTH__=$1
 local __STR__
 if [ -z "$__AUTH__" ]; then
  log_error "$FUNCNAME: authentication string is empty"
  return $ERR_LOGIN_FAILED
 fi
 __STR__=${__AUTH__%%:*}
 if [ -z "$__STR__" ]; then
  log_error "$FUNCNAME: empty string (user)"
  return $ERR_LOGIN_FAILED
 fi
 [[ $2 ]] && unset "$2" && eval $2=\$__STR__
 if [[ $3 ]]; then
  if [ "$2" = "$3" ]; then
   log_error "$FUNCNAME: user and password varname must not be the same"
  else
   __STR__=${__AUTH__#*:}
   if [ -z "$__STR__" -o "$__AUTH__" = "$__STR__" ]; then
    __STR__=$(prompt_for_password) || return $ERR_LOGIN_FAILED
   fi
   unset "$3" && eval $3=\$__STR__
  fi
 fi
}
list_submit() {
 local LINE I
 test "$1" || return $ERR_LINK_DEAD
 if test "$2"; then
  local -a LINKS NAMES
  mapfile -t LINKS <<< "$1"
  mapfile -t NAMES <<< "$2"
  if [[ "${#NAMES[@]}" -eq 1 ]]; then
   for I in "${!LINKS[@]}"; do
    test "${LINKS[$I]}" || continue
    echo "$3${LINKS[$I]}$4"
    echo "${NAMES[0]}"
   done
  else
   for I in "${!LINKS[@]}"; do
    test "${LINKS[$I]}" || continue
    echo "$3${LINKS[$I]}$4"
    echo "${NAMES[$I]}"
   done
  fi
 else
  while IFS= read -r LINE; do
   test "$LINE" || continue
   echo "$3$LINE$4"
   echo
  done <<< "$1"
 fi
}
translate_size() {
 local N=${1// }
 local S T
 N=${N//	}
 if [ -z "$N" ]; then
  log_error "$FUNCNAME: non empty argument expected"
  return $ERR_FATAL
 fi
 S=$(sed -ne '/[.,]/{s/^\(-\?[[:digit:]]*\)[.,]\([[:digit:]]\+\).*$/\1_\2/p;b;};
  s/^\(-\?[[:digit:]]\+\).*$/\1_/p' <<< "$N") || return $ERR_SYSTEM
 if [[ $S = '' || $S = '_' ]]; then
  log_error "$FUNCNAME: invalid parsed number \`$N'"
  return $ERR_FATAL
 fi
 local -i R=10#${S%_*}
 local -i F=0
 T=${S#*_}
 if test "$T"; then
  T="1${T}00"
  F=10#${T:1:3}
  T=$(( ${#S} ))
 else
  T=$(( ${#S} - 1 ))
 fi
 S=$(sed -e "s/^\.\?\([KkMmGg]i\?[Bb]\?\)$/\1/" <<< "${N:$T}") || return $ERR_SYSTEM
 case $S in
  k|kB)
   echo $(( 1000 * R + F))
   ;;
  KiB|Ki|K|KB)
   echo $(( 1024 * R + 1024 * F / 1000))
   ;;
  M|MB)
   echo $(( 1000000 * R + 1000 * F))
   ;;
  MiB|Mi|m|mB)
   echo $(( 1048576 * R + 1048576 * F / 1000))
   ;;
  G|GB)
   echo $(( 1000000000 * R + 1000000 * F))
   ;;
  GiB|Gi)
   echo $(( 1073741824 * R + 1073741824 * F / 1000))
   ;;
  B|'')
   echo "$R"
   ;;
  *b)
   log_error "$FUNCNAME: unknown unit \`$S' (we don't deal with bits, use B for bytes)"
   return $ERR_FATAL
   ;;
  *)
   log_error "$FUNCNAME: unknown unit \`$S'"
   return $ERR_FATAL
   ;;
 esac
}
storage_set() {
 local -r KEY=${1:-default}
 local CONFIG
 local -A OBJ
 if [ -z "$MODULE" ]; then
  log_error "$FUNCNAME: \$MODULE is undefined, abort"
  return $ERR_NOMODULE
 fi
 if [ "$CACHE" != 'shared' ]; then
  CONFIG="$TMPDIR/$(basename_file "$0").$$.${MODULE}.txt"
 else
  CONFIG="$PLOWSHARE_CONFDIR/storage"
  if [ ! -d "$CONFIG" ]; then
   mkdir -p "$CONFIG"
   chmod 700 "$CONFIG"
  fi
  if [ ! -w "$CONFIG" ]; then
    log_error "$FUNCNAME: write permissions expected \`$CONFIG'"
    return $ERR_SYSTEM
  fi
  CONFIG="$CONFIG/${MODULE}.txt"
 fi
 if [ -f "$CONFIG" ]; then
  if [ ! -w "$CONFIG" ]; then
   log_error "$FUNCNAME: write permissions expected \`$CONFIG'"
   return $ERR_SYSTEM
  fi
  source "$CONFIG"
 else
  touch "$CONFIG"
  chmod 600 "$CONFIG"
 fi
 if test "${2+isset}"; then
  OBJ[$KEY]=$2
 else
  unset -v OBJ[$KEY]
 fi
 declare -p OBJ >"$CONFIG"
 log_debug "$FUNCNAME: \`$KEY' set for module \`$MODULE'"
}
storage_get() {
 local -r KEY=${1:-default}
 local CONFIG
 local -A OBJ
 if [ -z "$MODULE" ]; then
  log_error "$FUNCNAME: \$MODULE is undefined, abort"
  return $ERR_NOMODULE
 fi
 if [ "$CACHE" != 'shared' ]; then
  CONFIG="$TMPDIR/$(basename_file "$0").$$.${MODULE}.txt"
 else
  CONFIG="$PLOWSHARE_CONFDIR/storage"
  [ -d "$CONFIG" ] || return $ERR_FATAL
  CONFIG="$CONFIG/${MODULE}.txt"
 fi
 if [ -f "$CONFIG" ]; then
  if [ ! -r "$CONFIG" ]; then
   log_error "$FUNCNAME: read permissions expected \`$CONFIG'"
   return $ERR_SYSTEM
  fi
  source "$CONFIG"
  if test "${OBJ[$KEY]+isset}"; then
   echo "${OBJ[$KEY]}"
   return 0
  fi
 fi
 return $ERR_FATAL
}
storage_reset() {
 local CONFIG
 if [ -z "$MODULE" ]; then
  log_error "$FUNCNAME: \$MODULE is undefined, abort"
  return $ERR_NOMODULE
 fi
 if [ "$CACHE" != 'shared' ]; then
  CONFIG="$TMPDIR/$(basename_file "$0").$$.${MODULE}.txt"
 else
  CONFIG="$PLOWSHARE_CONFDIR/storage"
  [ -d "$CONFIG" ] || return $ERR_FATAL
  CONFIG="$CONFIG/${MODULE}.txt"
 fi
 if [ -f "$CONFIG" ]; then
  rm -f "$CONFIG"
  log_debug "$FUNCNAME: delete file for module \`$MODULE'"
 fi
}
storage_timestamp_set() {
  storage_set '__date__' "$(date -u +%s)"
}
storage_timestamp_diff() {
  local CUR DATE
  DATE=$(storage_get '__date__') || return
  CUR=$(date -u +%s)
  [ -z "$1" ] || storage_set '__date__' "$CUR"
  echo "$((CUR - DATE))"
}
strip() {
 sed -e 's/\xC2\?\xA0/ /g' -e 's/^[[:space:]]*//; s/[[:space:]]*$//'
}
core_init() {
 local -r NAME=${1:-ERROR}
 if [ -z "$TMPDIR" ]; then
  log_error "$NAME: \$TMPDIR is undefined, abort"
  return $ERR_SYSTEM
 elif [ ! -d "$TMPDIR" ]; then
  log_error "$NAME: \$TMPDIR is not a directory, abort"
  return $ERR_SYSTEM
 fi
 if [ -n "$PLOWSHARE_CURL" ]; then
  if ! type -P "$PLOWSHARE_CURL" >/dev/null 2>&1; then
   log_error "$NAME: \$PLOWSHARE_CURL is invalid, abort"
   return $ERR_SYSTEM
  fi
  log_debug "using custom curl: $PLOWSHARE_CURL"
 fi
 if debian_javascript_enabled && [ -n "$PLOWSHARE_JS" ]; then
  if ! type -P "$PLOWSHARE_JS" >/dev/null 2>&1; then
   log_error "$NAME: \$PLOWSHARE_JS is invalid, abort"
   return $ERR_SYSTEM
  fi
  log_debug "using custom js: $PLOWSHARE_JS"
 fi
 trap 'log_notice_norc ""; rm -f "$TMPDIR/$(basename_file $0).$$".*' EXIT TERM
 trap 'CONT_SIGNAL=$(date +%s)' CONT
}
check_exec() {
 command -v "$1" >/dev/null 2>&1
}
timeout_init() {
 PS_TIMEOUT=$1
}
print_options() {
 local -r INDENT=${2:-'  '}
 local STR VAR SHORT LONG TYPE MSG
 while read -r; do
  test "$REPLY" || continue
  IFS=',' read -r VAR SHORT LONG TYPE MSG <<< "$REPLY"
  if [ -n "$SHORT" ]; then
   if test "$TYPE"; then
    STR="-${SHORT} ${TYPE#*=}"
    test -n "$LONG" && STR="-${SHORT}, --${LONG}=${TYPE#*=}"
   else
    STR="-${SHORT}"
    test -n "$LONG" && STR="$STR, --${LONG}"
   fi
  else
   if test "$TYPE"; then
    STR="    --${LONG}=${TYPE#*=}"
   else
    STR="    --${LONG}"
   fi
  fi
  printf '%-35s%s\n' "$INDENT$STR" "$MSG"
 done <<< "$1"
}
print_module_options() {
 local ELT OPTIONS
 for ELT in "${!1}"; do
  OPTIONS=$(get_module_options "$ELT" "$2")
  if test "$OPTIONS"; then
   echo
   echo "Options for module <$ELT>:"
   print_options "$OPTIONS"
  fi
 done
}
get_all_modules_options() {
 local ELT
 for ELT in "${!1}"; do
  get_module_options "$ELT" "$2"
 done
}
get_module() {
 local ELT
 for ELT in "${!2}"; do
  local -u VAR="MODULE_${ELT}_REGEXP_URL"
  if match "${!VAR}" "$1"; then
   echo "$ELT"
   return 0
  fi
 done
 return $ERR_NOMODULE
}
process_core_options() {
 local -r NAME=$1
 local -r OPTIONS=$(strip_and_drop_empty_lines "$2")
 shift 2
 if [ -d "$PLOWSHARE_CONFDIR/exec" ]; then
  VERBOSE=2 PATH="$PLOWSHARE_CONFDIR/exec:$PATH" process_options \
   "$NAME" "$OPTIONS" -1 "$@"
 else
  VERBOSE=2 process_options "$NAME" "$OPTIONS" -1 "$@"
 fi
}
process_all_modules_options() {
 local -r NAME=$1
 local -r OPTIONS=$2
 shift 2
 process_options "$NAME" "$OPTIONS" 0 "$@"
}
process_module_options() {
 local -r M=$1
 local -r OPTIONS=$(get_module_options "$1" "$2")
 shift 2
 process_options "$M" "$OPTIONS" 1 "$@"
}
get_all_modules_list() {
 local -a SRCS=( "$LIBDIR/modules" "$PLOWSHARE_CONFDIR/modules" )
 local -A MODULES_PATH=()
 local D CONFIG
 if [ -d "$PLOWSHARE_CONFDIR/modules.d/" ]; then
  while read -r; do
   D=$(dirname "$REPLY")
   SRCS+=( "$D" )
  done < <(find -L "$PLOWSHARE_CONFDIR/modules.d/" -mindepth 2 -maxdepth 2 -name config)
 fi
 for D in "${SRCS[@]}"; do
  CONFIG="$D/config"
  if [[ -d "$D" && -f "$CONFIG" ]]; then
   while read -r; do
    if [ -f "$D/$REPLY.sh" ]; then
     MODULES_PATH[$REPLY]="$D/$REPLY.sh"
    else
     stderr "ERROR: $CONFIG: \`$REPLY\` module not found, ignoring"
    fi
   done < <(if test "$2"; then sed -ne \
    "/^[^#]/{/|[[:space:]]*$1/{/|[[:space:]]*$2/!s/^\([^[:space:]|]*\).*/\1/p}}" \
    "$CONFIG"; else sed -ne \
    "/^[^#]/{/|[[:space:]]*$1/s/^\([^[:space:]|]*\).*/\1/p}" "$CONFIG"; fi)
  fi
 done
 declare -p MODULES_PATH
}
process_configfile_options() {
 local CONFIG OPTIONS SECTION NAME VALUE OPTION
 if [ -z "$3" ]; then
  CONFIG="$PLOWSHARE_CONFDIR/plowshare.conf"
  test -r "$CONFIG" || CONFIG='/etc/plowshare.conf'
  test -r "$CONFIG" || return 0
 else
  CONFIG=$3
 fi
 OPTIONS=$(strip_and_drop_empty_lines "$2")
 SECTION=$(sed -ne "/\[$1\]/,/^\[/H; /\[[Gg]eneral\]/,/^\[/p; \${x;p}" \
  "$CONFIG" | sed -e '/^\(#\|\[\|[[:space:]]*$\)/d') || true
 if [ -n "$SECTION" -a -n "$OPTIONS" ]; then
  while read -r; do
   NAME=$(strip <<< "${REPLY%%=*}")
   VALUE=$(strip <<< "${REPLY#*=}")
   [[ $NAME = */* ]] && continue
   if [ ${#VALUE} -gt 1 ] && [ '"' = "${VALUE:0:1}" -a '"' = "${VALUE:(-1):1}" ]; then
    VALUE=${VALUE%?}
    VALUE=${VALUE:1}
   fi
   OPTION=$(sed -ne "/,.\?,${NAME},/{p;q}" <<< "$OPTIONS") || true
   if [ -n "$OPTION" ]; then
    local VAR=${OPTION%%,*}
    eval "$VAR=\$VALUE"
   fi
  done <<< "$SECTION"
 fi
}
process_configfile_module_options() {
 local CONFIG OPTIONS SECTION OPTION LINE VALUE
 if [ -z "$4" ]; then
  CONFIG="$PLOWSHARE_CONFDIR/plowshare.conf"
  test -r "$CONFIG" || CONFIG='/etc/plowshare.conf'
  test -r "$CONFIG" || return 0
 else
  CONFIG=$4
 fi
 if [ -r "$CONFIG" ]; then
  if [ "$CONFIG" != '/etc/plowshare.conf' ]; then
   if [ -O "$CONFIG" ]; then
    local FILE_PERM=$(ls -l -L "$CONFIG" 2>/dev/null)
    if [[ ${FILE_PERM:4:6} != '------' ]]; then
     log_notice "WARNING: Wrong configuration file permissions. Fix it with: chmod 600 $CONFIG"
    fi
   else
    log_notice "WARNING: Bad configuration file ownership. Fix it with: chown $USER $CONFIG"
   fi
  fi
 else
  log_error "ERROR: cannot access file \`$CONFIG'"
  return 0
 fi
 log_report "use $CONFIG"
 OPTIONS=$(get_module_options "$2" "$3")
 SECTION=$(sed -ne "/\[[Gg]eneral\]/,/^\[/H; /\[$1\]/,/^\[/p; \${x;p}" \
  "$CONFIG" | sed -e '/^\(#\|\[\|[[:space:]]*$\)/d') || true
 if [ -n "$SECTION" -a -n "$OPTIONS" ]; then
  local VAR SHORT LONG
  local -lr M=$2
  while read -r; do
   IFS=',' read -r VAR SHORT LONG _ <<< "$REPLY"
   LINE=$(sed -ne "/^[[:space:]]*$M\/\($SHORT\|$LONG\)[[:space:]]*=/{p;q}" <<< "$SECTION") || true
   if [ -n "$LINE" ]; then
    VALUE=$(strip <<< "${LINE#*=}")
    if [ ${#VALUE} -gt 1 ] && [ '"' = "${VALUE:0:1}" -a '"' = "${VALUE:(-1):1}" ]; then
     VALUE=${VALUE%?}
     VALUE=${VALUE:1}
    fi
    eval "$VAR=\$VALUE"
    log_notice "$M: take --$LONG option from configuration file"
   else
    unset "$VAR"
   fi
  done <<< "$OPTIONS"
 fi
}
log_report_info() {
 local -r LIBDIR1=$1
 local G LIBDIR2
 if test $VERBOSE -ge 4; then
  log_report '=== SYSTEM INFO BEGIN ==='
  log_report "[mach] $HOSTNAME $HOSTTYPE $OSTYPE $MACHTYPE"
  log_report "[bash] $BASH_VERSION"
  test "$http_proxy" && log_report "[env ] http_proxy=$http_proxy"
  if check_exec 'curl'; then
   log_report "[curl] $(command curl --version | first_line)"
   [ -f "$HOME/.curlrc" ] && \
    log_report '[curl] ~/.curlrc exists'
  else
   log_report '[curl] not found!'
  fi
  check_exec 'gsed' && G=g
  log_report "[sed ] $(${G}sed --version | sed -ne '/version/p')"
  log_report "[lib ] '$LIBDIR1'"
  for LIBDIR2 in '/usr/share/plowshare' '/usr/local/share/plowshare'; do
   if [ "$LIBDIR2" != "$LIBDIR1" -a -f "$LIBDIR2/core.sh" ]; then
    log_report "[lib2] '$LIBDIR2'"
   fi
  done
  if git -C "$LIBDIR" rev-parse --is-inside-work-tree &>/dev/null; then
   local -r GIT_BRANCH=$(git -C "$LIBDIR" rev-parse --quiet --abbrev-ref HEAD)
   local -r GIT_REV=$(git -C "$LIBDIR" describe --tags --always 2>/dev/null)
   log_report "[git ] $GIT_REV ($GIT_BRANCH branch)"
  fi
  log_report '=== SYSTEM INFO END ==='
 fi
}
captcha_method_translate() {
 case $1 in
  none)
   [[ $2 ]] && unset "$2" && eval $2=none
   [[ $3 ]] && unset "$3" && eval $3=log
   ;;
  imgur)
   [[ $2 ]] && unset "$2" && eval $2=prompt
   [[ $3 ]] && unset "$3" && eval $3=imgur
   ;;
  x11|xorg)
   if [ -z "$DISPLAY" ]; then
    log_error 'Cannot open display. Are you running on a X11/Xorg environment ?'
    return $ERR_FATAL
   fi
   [[ $2 ]] && unset "$2" && eval $2=prompt
   [[ $3 ]] && unset "$3" && eval $3=view-x,log
   ;;
  nox)
   [[ $2 ]] && unset "$2" && eval $2=prompt
   [[ $3 ]] && unset "$3" && eval $3=view-aa,log
   ;;
  online)
   if [ -z "$CAPTCHA_ANTIGATE" -a -z "$CAPTCHA_9KWEU" -a \
     -z "$CAPTCHA_BHOOD" -a -z "$CAPTCHA_COIN" -a -z "$CAPTCHA_DEATHBY" ]; then
    log_error 'No captcha solver account provided. Consider using --9kweu, --antigate, --captchabhood, --captchacoin or --deathbycaptcha options.'
    return $ERR_FATAL
   fi
   [[ $2 ]] && unset "$2" && eval $2=online
   [[ $3 ]] && unset "$3" && eval $3=log
   ;;
  fb|fb0)
   if ! test -c '/dev/fb0'; then
    log_error 'Cannot find framebuffer device /dev/fb0'
    return $ERR_FATAL
   fi
   [[ $2 ]] && unset "$2" && eval $2=prompt
   [[ $3 ]] && unset "$3" && eval $3=view-fb,log
   ;;
  *)
   log_error "ERROR: Unknown captcha method '$1'.${DISPLAY:+ Try with 'x11' for example.}"
   return $ERR_FATAL
   ;;
 esac
 return 0
}
handle_tokens() {
 local S=$1
 local -A FMT
 local OUT K REGEX
 shift
 for K in "$@"; do
  REGEX+="${K%%,*}|"
  FMT[${K%%,*}]=${K#*,}
 done
 REGEX=${REGEX%|}
 OUT='x'
 while [[ $S =~ ^(.*)($REGEX)(.*)$ ]]; do
  S=${BASH_REMATCH[1]}
  OUT="${FMT[${BASH_REMATCH[2]}]}${BASH_REMATCH[3]}$OUT"
 done
 OUT="$S$OUT"
 echo -n "${OUT%x}"
}
json_escape() {
 local S=${1//\\/\\\\}
 S=${S//\//\\/}
 echo -n "${S//\"/\\\"}"
}
stderr() {
 echo "$@" >&2
 return 0
}
quote() {
 echo \'${1//\'/\'\\\'\'}\' #'# Help vim syntax highlighting
}
quote_multiple() {
  if [[ $1 ]]; then
   echo "${1%)}"
  else
   echo '('
  fi
  quote "$2"
  echo ')'
}
quote_array() {
 local -a ARR
 local E
 IFS=',' read -r -a ARR <<< "$1"
 echo '('
 for E in "${ARR[@]}"; do
  quote "$(strip <<< "$E")"
 done
 echo ')'
}
translate_exec() {
 local F=$1
 if test -x "$F"; then
  [[ $F = /* ]] || F="$PWD/$F"
 else
  F=$(type -P "$F" 2>/dev/null)
 fi
 if [ -z "$F" ]; then
  log_error "$FUNCNAME: failed ($1)"
  return $ERR_FATAL
 fi
 quote "$F"
}
check_transfer_speed() {
 local N=${1// }
 case $N in
  *Ki|*Mi)
   N=${N%??}
   ;;
  *K|*m|*k|*M)
   N=${N%?}
   ;;
  *)
   ;;
 esac
 if [[ $N = *[![:digit:]]* || $N -eq 0 ]]; then
  return 1
 fi
}
check_disk_size() {
 local N=${1// }
 case $N in
  *Mi|*Gi|*MB|*GB)
   N=${N%??}
   ;;
  *M|*m|*G)
   N=${N%?}
   ;;
  *)
   N=err
   ;;
 esac
 if [[ $N = *[![:digit:]]* || $N -eq 0 ]]; then
  return 1
 fi
}
grep_block_by_order() {
 local -r TAG=$1
 local DATA=$2
 local N=${3:-'1'}
 local MAX NEW
 MAX=$(grep -c "<$TAG[[:space:]>]" <<< "$DATA") || true
 if (( $N < 0 )); then
  N=$(( $MAX + 1 + N ))
  if (( $N <= 0 )); then
   log_error "${FUNCNAME[1]} failed: negative index is too big (detected $MAX forms)"
   return $ERR_FATAL
  fi
 fi
 NEW=${TAG//[}
 NEW=${NEW//]}
 while [ "$N" -gt "1" ]; do
  (( --N ))
  DATA=$(echo "$DATA" | sed -ne "/<\/$TAG>/,\$p" | \
   sed -e "1s/<\/\?$TAG[[:space:]>]/<_${NEW}_>/g")
  test -z "$DATA" && break
 done
 local STR=$(sed -ne \
  "/<$TAG[[:space:]>]/,/<\/$TAG>/{p;/<\/$TAG/q}" <<< "$DATA")
 if [ -z "$STR" ]; then
  log_error "${FUNCNAME[1]} failed (sed): \"n=$N\""
  return $ERR_FATAL
 fi
 echo "$STR"
}
check_argument_type() {
 local -r NAME=$1
 local -r TYPE=$2
 local -r VAL=$3
 local -r OPT=$4
 local RET=$ERR_BAD_COMMAND_LINE
 if [[ $TYPE = 'a' && $VAL != *:* ]]; then
  log_debug "$NAME ($OPT): missing password for credentials"
  RET=0
 elif [[ $TYPE = 'n' && ( $VAL = *[![:digit:]]* || $VAL -le 0 ) ]]; then
  log_error "$NAME ($OPT): positive integer expected"
 elif [[ $TYPE = 'N' && ( $VAL = *[![:digit:]]* || $VAL = '' ) ]]; then
  log_error "$NAME ($OPT): positive or zero integer expected"
 elif [[ $TYPE = [st] && $VAL = '' ]]; then
  log_error "$NAME ($OPT): empty string not expected"
 elif [ "$TYPE" = 'r' ] && ! check_transfer_speed "$VAL"; then
  log_error "$NAME ($OPT): positive transfer rate expected"
 elif [ "$TYPE" = 'R' ] && ! check_disk_size "$VAL"; then
  log_error "$NAME ($OPT): wrong value, megabyte or gigabyte suffix is mandatory"
 elif [[ $TYPE = 'e' && "${VAL#*@*.}" = "$VAL" ]]; then
  log_error "$NAME ($OPT): invalid email address"
 elif [ $TYPE = 'f' ]; then
  if test -z "$VAL"; then
   log_error "$NAME ($OPT): filename expected"
  elif test -f "$VAL"; then
   if test -r "$VAL"; then
    RET=0
   else
    log_error "$NAME ($OPT): read permissions expected \`$VAL'"
   fi
  else
   log_error "$NAME ($OPT): cannot access file \`$VAL'"
  fi
 elif [ $TYPE = 'F' ]; then
  if test -z "$VAL"; then
   log_error "$NAME ($OPT): executable filename expected"
  elif test -f "$VAL"; then
   if test -x "$VAL"; then
    RET=0
   else
    log_error "$NAME ($OPT): executable permissions expected \`$VAL'"
   fi
  elif check_exec "$VAL"; then
   local F=$(type -P "$VAL" 2>/dev/null)
   if test -x "$F"; then
    RET=0
   else
    log_error "$NAME ($OPT): executable permissions expected \`$F'"
   fi
  else
   log_error "$NAME ($OPT): cannot access file \`$VAL'"
  fi
 elif [ $TYPE = 'D' ]; then
  if test -z "$VAL"; then
   log_error "$NAME ($OPT): directory expected"
  elif test -d "$VAL"; then
   if test -w "$VAL"; then
    RET=0
   else
    log_error "$NAME ($OPT): write permissions expected \`$VAL'"
   fi
  else
   log_error "$NAME ($OPT): cannot access directory \`$VAL'"
  fi
 elif [[ $TYPE = 'l' && $VAL = '' ]]; then
  log_error "$NAME ($OPT): comma-separated list expected"
 elif [[ $TYPE = [cC]* ]]; then
  if [[ $TYPE = C* && $VAL = '' ]]; then
   RET=0
  else
   local -a ITEMS
   IFS='|' read -r -a ITEMS <<< "${TYPE:2}"
   if find_in_array ITEMS[@] "$VAL"; then
    RET=0
   else
    log_error "$NAME ($OPT): wrong value '$VAL'. Possible values are: ${ITEMS[*]}."
   fi
  fi
 elif [[ $TYPE = [lsSt] ]]; then
  RET=0
 elif [[ $TYPE = [aenNrR] ]]; then
  if [ "${VAL:0:1}" = '-' ]; then
   log_error "$NAME ($OPT): missing parameter"
  else
   RET=0
  fi
 else
  log_error "$NAME: unknown argument type '$TYPE'"
 fi
 [ $RET -eq 0 ] || echo false
 return $RET
}
process_options() {
 local -r NAME=$1
 local -r OPTIONS=$2
 local -r STEP=$3
 local -A RES
 local -a UNUSED_OPTS UNUSED_ARGS
 local -a OPTS_VAR_LONG OPTS_NAME_LONG OPTS_TYPE_LONG
 local -a OPTS_VAR_SHORT OPTS_NAME_SHORT OPTS_TYPE_SHORT
 local ARG VAR SHORT LONG TYPE SKIP_ARG FOUND FUNC
 shift 3
 UNUSED_OPTS=()
 UNUSED_ARGS=()
 if [ -z "$OPTIONS" ]; then
  if [ $STEP -gt 0 ]; then
   echo "${NAME}_vars_set() { :; }"
   echo "${NAME}_vars_unset() { :; }"
   return 0
  fi
 else
  while read -r ARG; do
   IFS=',' read -r VAR SHORT LONG TYPE _ <<< "$ARG"
   if [ -n "$LONG" ]; then
    OPTS_VAR_LONG[${#OPTS_VAR_LONG[@]}]=$VAR
    OPTS_NAME_LONG[${#OPTS_NAME_LONG[@]}]="--$LONG"
    OPTS_TYPE_LONG[${#OPTS_TYPE_LONG[@]}]=$TYPE
   fi
   if [ -n "$SHORT" ]; then
    OPTS_VAR_SHORT[${#OPTS_VAR_SHORT[@]}]=$VAR
    OPTS_NAME_SHORT[${#OPTS_NAME_SHORT[@]}]="-$SHORT"
    OPTS_TYPE_SHORT[${#OPTS_TYPE_SHORT[@]}]=$TYPE
   fi
  done <<< "$OPTIONS"
 fi
 for ARG in "$@"; do
  shift
  if [ -n "$SKIP_ARG" ]; then
   unset SKIP_ARG
   [ $STEP -eq 0 ] && UNUSED_OPTS[${#UNUSED_OPTS[@]}]="$ARG"
   continue
  fi
  if [ "$ARG" = '--' ]; then
   UNUSED_ARGS=("${UNUSED_ARGS[@]}" "$@")
   break
  fi
  unset FOUND
  if [ "${ARG:0:2}" = '--' ]; then
   for I in "${!OPTS_NAME_LONG[@]}"; do
    if [ "${OPTS_NAME_LONG[$I]}" = "${ARG%%=*}" ]; then
     TYPE=${OPTS_TYPE_LONG[$I]%%=*}
     if [ "$TYPE" = 'l' ]; then
      FUNC=quote_array
     elif [[ $TYPE = [rR] ]]; then
      FUNC=translate_size
     elif [ "$TYPE" = 'F' ]; then
      FUNC=translate_exec
     elif [ "$TYPE" = 't' ]; then
      FUNC=quote_multiple
     else
      FUNC=quote
     fi
     if [ -z "$TYPE" ]; then
      [[ ${RES[${OPTS_VAR_LONG[$I]}]} ]] && \
       log_notice "$NAME: useless duplicated option ${ARG%%=*}, ignoring"
      RES[${OPTS_VAR_LONG[$I]}]=1
      [ "${ARG%%=*}" != "$ARG" ] && \
       log_notice "$NAME: unwanted argument \`${ARG#*=}' for ${ARG%%=*}, ignoring"
     elif [ "${ARG%%=*}" != "$ARG" ]; then
      [ $STEP -gt 0 ] || check_argument_type "$NAME" \
       "$TYPE" "${ARG#*=}" "${ARG%%=*}" || return
      if [ "$TYPE" = 't' ]; then
       RES[${OPTS_VAR_LONG[$I]}]="$($FUNC "${RES[${OPTS_VAR_LONG[$I]}]}" "${ARG#*=}")"
      else
       [[ ${RES[${OPTS_VAR_LONG[$I]}]} ]] && \
        log_notice "$NAME: duplicated option ${ARG%%=*}, taking last one"
       RES[${OPTS_VAR_LONG[$I]}]="$($FUNC "${ARG#*=}")"
      fi
     else
      if [ $# -eq 0 ]; then
       log_error "$NAME ($ARG): argument required"
       echo "exit $ERR_BAD_COMMAND_LINE"
       return $ERR_BAD_COMMAND_LINE
      fi
      [ $STEP -gt 0 ] || check_argument_type "$NAME" \
       "$TYPE" "$1" "$ARG" || return
      if [ "$TYPE" = 't' ]; then
       RES[${OPTS_VAR_LONG[$I]}]="$($FUNC "${RES[${OPTS_VAR_LONG[$I]}]}" "$1")"
      else
       [[ ${RES[${OPTS_VAR_LONG[$I]}]} ]] && \
        log_notice "$NAME: duplicated option $ARG, taking last one"
       RES[${OPTS_VAR_LONG[$I]}]="$($FUNC "$1")"
      fi
      SKIP_ARG=1
     fi
     FOUND=1
     break
    fi
   done
  elif [ "${ARG:0:1}" = '-' ]; then
   for I in "${!OPTS_NAME_SHORT[@]}"; do
    if [ "${OPTS_NAME_SHORT[$I]}" = "${ARG:0:2}" ]; then
     TYPE=${OPTS_TYPE_SHORT[$I]%%=*}
     if [ "$TYPE" = 'l' ]; then
      FUNC=quote_array
     elif [[ $TYPE = [rR] ]]; then
      FUNC=translate_size
     elif [ "$TYPE" = 'F' ]; then
      FUNC=translate_exec
     elif [ "$TYPE" = 't' ]; then
      FUNC=quote_multiple
     else
      FUNC=quote
     fi
     if [ -z "$TYPE" ]; then
      [[ ${RES[${OPTS_VAR_SHORT[$I]}]} ]] && \
       log_notice "$NAME: useless duplicated option ${ARG:0:2}, ignoring"
      RES[${OPTS_VAR_SHORT[$I]}]=1
      [[ ${#ARG} -gt 2 ]] && \
       log_notice "$NAME: unwanted argument \`${ARG:2}' for ${ARG:0:2}, ignoring"
     elif [ ${#ARG} -gt 2 ]; then
      [ $STEP -gt 0 ] || check_argument_type "$NAME" \
       "$TYPE" "${ARG:2}" "${ARG:0:2}" || return
      if [ "$TYPE" = 't' ]; then
       RES[${OPTS_VAR_SHORT[$I]}]="$($FUNC "${RES[${OPTS_VAR_SHORT[$I]}]}" "${ARG:2}")"
      else
       [[ ${RES[${OPTS_VAR_SHORT[$I]}]} ]] && \
        log_notice "$NAME: duplicated option ${ARG:0:2}, taking last one"
       RES[${OPTS_VAR_SHORT[$I]}]="$($FUNC "${ARG:2}")"
      fi
     else
      if [ $# -eq 0 ]; then
       log_error "$NAME ($ARG): argument required"
       echo "exit $ERR_BAD_COMMAND_LINE"
       return $ERR_BAD_COMMAND_LINE
      fi
      [ $STEP -gt 0 ] || check_argument_type "$NAME" \
       "$TYPE" "$1" "$ARG" || return
      if [ "$TYPE" = 't' ]; then
       RES[${OPTS_VAR_SHORT[$I]}]="$($FUNC "${RES[${OPTS_VAR_SHORT[$I]}]}" "$1")"
      else
       [[ ${RES[${OPTS_VAR_SHORT[$I]}]} ]] && \
        log_notice "$NAME: duplicated option $ARG, taking last one"
       RES[${OPTS_VAR_SHORT[$I]}]="$($FUNC "$1")"
      fi
      SKIP_ARG=1
     fi
     FOUND=1
     break
    fi
   done
  fi
  if [ -z "$FOUND" ]; then
   if [[ $ARG =~ ^-[[:alnum:]][[:alnum:]-]+ ]]; then
    if find_in_array OPTS_NAME_LONG[@] "-${BASH_REMATCH[0]}"; then
     log_error "$NAME: unknown command-line option \`$ARG\`, do you mean \`-${BASH_REMATCH[0]}\` (with two dashes)?"
     echo "exit $ERR_BAD_COMMAND_LINE"
     return $ERR_BAD_COMMAND_LINE
    fi
   fi
   if [ $STEP -lt 0 ]; then
    [[ $ARG = --??* ]] && \
     CLOSE_OPT=$(close_match_in_array OPTS_NAME_LONG[@] "$ARG") && \
      declare -p CLOSE_OPT
   fi
   if [ $STEP -eq 0 ]; then
    if [[ $ARG = -?* ]]; then
     if [[ $CLOSE_OPT ]]; then
      log_error "$NAME: unknown command-line option \`$ARG\`, do you mean \`$CLOSE_OPT\` (close match)?"
     else
      log_error "$NAME: unknown command-line option \`$ARG\`"
     fi
     echo "exit $ERR_BAD_COMMAND_LINE"
     return $ERR_BAD_COMMAND_LINE
    fi
    UNUSED_ARGS[${#UNUSED_ARGS[@]}]="$ARG"
   else
    UNUSED_OPTS[${#UNUSED_OPTS[@]}]="$ARG"
   fi
  elif [ $STEP -eq 0 ]; then
   UNUSED_OPTS[${#UNUSED_OPTS[@]}]="$ARG"
  fi
 done
 if [ $STEP -lt 0 ]; then
  for ARG in "${!RES[@]}"; do echo "declare -r ${ARG}=${RES[$ARG]}"; done
 elif [ $STEP -gt 0 ]; then
  echo "${NAME}_vars_set() { :"
  for ARG in "${!RES[@]}"; do echo "${ARG}=${RES[$ARG]}"; done
  echo '}'
  echo "${NAME}_vars_unset() { :"
  for ARG in "${!RES[@]}"; do echo "${ARG%%=*}="; done
  echo '}'
 fi
 declare -p UNUSED_ARGS
 declare -p UNUSED_OPTS
}
strip_and_drop_empty_lines() {
 sed -e '/^[[:space:]]*$/d; s/^[[:space:]]*//; s/[[:space:]]*$//' <<< "$1"
}
get_module_options() {
 local -ur VAR="MODULE_${1}_${2}_OPTIONS"
 strip_and_drop_empty_lines "${!VAR}"
}
splitseconds() {
 local DIV_H=$(( $1 / 3600 ))
 local DIV_M=$(( ($1 % 3600) / 60 ))
 local DIV_S=$(( $1 % 60 ))
 [ "$DIV_H" -eq 0 ] || echo -n "${DIV_H}h"
 [ "$DIV_M" -eq 0 ] || echo -n "${DIV_M}m"
 [ "$DIV_S" -eq 0 ] && echo || echo "${DIV_S}s"
}
timeout_update() {
 local WAIT=$1
 test -z "$PS_TIMEOUT" && return
 log_debug "time left to timeout: $PS_TIMEOUT secs"
 if [[ $PS_TIMEOUT -lt $WAIT ]]; then
  log_notice "Timeout reached (asked to wait $WAIT seconds, but remaining time is $PS_TIMEOUT)"
  return $ERR_MAX_WAIT_REACHED
 fi
 (( PS_TIMEOUT -= WAIT ))
}
find_in_array() {
 local ELT
 for ELT in "${!1}"; do
  [ "$ELT" = "$2" -o "$ELT" = "$3" ] && return 0
 done
 return 1
}
close_match_in_array() {
 local ELT
 local S=${2/#--no/--no-?}
 for ELT in "${!1}"; do
  if [[ $ELT =~ ^$S.?$ ]]; then
   echo "$ELT"
   return 0
  fi
 done
 return 1
}
index_in_array() {
 local ELT I=0
 for ELT in "${!1}"; do
  (( ++I ))
  if [ "$ELT" = "$2" -o "$ELT" = "$3" ]; then
   echo "$I"
   return 0
  fi
 done
 return 1
}
service_9kweu_ready() {
 local -r KEY=$1
 local AMOUNT
 if [ -z "$KEY" ]; then
  return $ERR_FATAL
 fi
 AMOUNT=$(curl --get --data 'action=usercaptchaguthaben' \
  --data "apikey=$CAPTCHA_9KWEU" 'http://www.9kw.eu/index.cgi') || { \
  log_notice '9kweu: site seems to be down'
  return $ERR_NETWORK
 }
 if [[ $AMOUNT = 00[012][[:digit:]][[:space:]]* ]]; then
  if [ "${AMOUNT:0:5}" = '0011 ' ]; then
   log_notice '9kw.eu: no more credits'
  else
   log_error "9kw.eu remote error: ${AMOUNT:5}"
  fi
 elif (( AMOUNT < 10 )); then
  log_notice '9kw.eu: insufficient credits'
 else
  log_debug "9kw.eu credits: $AMOUNT"
  return 0
 fi
 return $ERR_FATAL
}
service_antigate_ready() {
 local -r KEY=$1
 local AMOUNT LOAD
 if [ -z "$KEY" ]; then
  return $ERR_FATAL
 fi
 AMOUNT=$(curl --get --data "key=${CAPTCHA_ANTIGATE}&action=getbalance" \
  'http://antigate.com/res.php') || { \
  log_notice 'antigate: site seems to be down'
  return $ERR_NETWORK
 }
 if match '500 Internal Server Error' "$AMOUNT"; then
  log_error 'antigate: internal server error (HTTP 500)'
  return $ERR_CAPTCHA
 elif match '502 Bad Gateway' "$AMOUNT"; then
  log_error 'antigate: bad gateway (HTTP 502)'
  return $ERR_CAPTCHA
 elif match '503 Service Unavailable' "$AMOUNT"; then
  log_error 'antigate: service unavailable (HTTP 503)'
  return $ERR_CAPTCHA
 elif match '^ERROR' "$AMOUNT"; then
  log_error "antigate error: $AMOUNT"
  return $ERR_FATAL
 elif [ '0.0000' = "$AMOUNT" -o '-' = "${AMOUNT:0:1}" ]; then
  log_notice 'antigate: no more credits (or bad key)'
  return $ERR_FATAL
 else
  log_debug "antigate credits: \$$AMOUNT"
 fi
 if LOAD=$(curl 'http://antigate.com/load.php'); then
  local N P B T
  N=$(parse_tag 'waiting' <<< "$LOAD")
  P=$(parse_tag 'load' <<< "$LOAD")
  B=$(parse_tag 'minbid' <<< "$LOAD")
  T=$(parse_tag 'averageRecognitionTime' <<< "$LOAD")
  log_debug "Available workers: $N (load: ${P}%)"
  log_debug "Mininum bid: $B"
  log_debug "Average recognition time (s): $T"
  if [ "$N" = '0' -a "$P" = '100' ]; then
   log_notice 'antigate: no slot available (all workers are busy)'
   return $ERR_FATAL
  fi
 fi
}
service_captchabrotherhood_ready() {
 local RESPONSE AMOUNT ERROR
 if [ -z "$1" -o -z "$2" ]; then
  return $ERR_FATAL
 fi
 RESPONSE=$(curl --get -d "username=$1" -d "password=$2" \
  'http://www.captchabrotherhood.com/askCredits.aspx') || return
 if [ "${RESPONSE:0:3}" = 'OK-' ]; then
  AMOUNT=${RESPONSE:3}
  if (( AMOUNT < 10 )); then
   log_notice "CaptchaBrotherHood: not enough credits ($1)"
   return $ERR_FATAL
  fi
 else
  ERROR=${RESPONSE#Error-}
  log_error "CaptchaBrotherHood error: $ERROR"
  return $ERR_FATAL
 fi
 log_debug "CaptchaBrotherhood credits: $AMOUNT"
}
service_captchacoin_ready() {
 local -r KEY=$1
 local JSON ERROR AMOUNT
 if [ -z "$KEY" ]; then
  return $ERR_FATAL
 fi
 JSON=$(curl --get -d "api_key=${KEY}" \
   'http://api.captchacoin.com/api/details') || { \
  log_notice 'CaptchaCoin: site seems to be down'
  return $ERR_NETWORK
 }
 if match_json_true 'success' "$JSON"; then
  AMOUNT=$(echo "$JSON" | parse_json 'balance')
  if (( AMOUNT < 100 )); then
   log_error 'CaptchaCoin: not enough credits'
   return $ERR_FATAL
  fi
 else
  ERROR=$(echo "$JSON" | parse_json_quiet 'error')
  if [ -n "$ERROR" ]; then
   log_error "CaptchaCoin error: $ERROR"
  else
   log_error "CaptchaCoin unknown error: $JSON"
  fi
  return $ERR_FATAL
 fi
 log_debug "CaptchaCoin credits: $AMOUNT"
}
service_captchadeathby_ready() {
 local -r USER=$1
 local JSON STATUS AMOUNT ERROR
 if [ -z "$1" -o -z "$2" ]; then
  return $ERR_FATAL
 fi
 JSON=$(curl -F "username=$USER" -F "password=$2" \
   --header 'Accept: application/json' \
   'http://api.dbcapi.me/api/user') || { \
  log_notice 'DeathByCaptcha: site seems to be down'
  return $ERR_NETWORK
 }
 STATUS=$(echo "$JSON" | parse_json_quiet 'status')
 if [ "$STATUS" = 0 ]; then
  AMOUNT=$(echo "$JSON" | parse_json 'balance')
  if match_json_true 'is_banned' "$JSON"; then
   log_error "DeathByCaptcha error: $USER is banned"
   return $ERR_FATAL
  fi
  if [ "${AMOUNT%.*}" = 0 ]; then
   log_notice "DeathByCaptcha: not enough credits ($USER)"
   return $ERR_FATAL
  fi
 elif [ "$STATUS" = 255 ]; then
  ERROR=$(echo "$JSON" | parse_json_quiet 'error')
  log_error "DeathByCaptcha error: $ERROR"
  return $ERR_FATAL
 else
  log_error "DeathByCaptcha unknown error: $JSON"
  return $ERR_FATAL
 fi
 log_debug "DeathByCaptcha credits: $AMOUNT"
}
image_upload_imgur() {
 local -r IMG=$1
 local -r BASE_API='https://api.imgur.com/3'
 local RESPONSE DIRECT_URL FID DEL_HASH
 log_debug 'uploading image to Imgur.com'
 RESPONSE=$(curl -F "image=@$IMG" \
  -H 'Authorization: Client-ID 5926b561daf4510' \
  --form-string 'type=file' \
  --form-string 'title=Plowshare uploaded image' \
  "$BASE_API/upload.json") || return
 DIRECT_URL=$(parse_json_quiet link <<< "$RESPONSE")
 FID=$(parse_json_quiet id <<< "$RESPONSE")
 DEL_HASH=$(parse_json_quiet deletehash <<< "$RESPONSE")
 if [ -z "$DIRECT_URL" -o -z "$FID" ]; then
  log_debug "$FUNCNAME: $RESPONSE"
  if match '504 Gateway Time-out' "$RESPONSE"; then
   log_error "$FUNCNAME: upload error (Gateway Time-out)"
  elif match 'Imgur is over capacity' "$RESPONSE"; then
   log_error "$FUNCNAME: upload error (Service Unavailable)"
  else
   log_error "$FUNCNAME: upload error"
  fi
  return $ERR_FATAL
 fi
 log_error "Image: $DIRECT_URL"
 log_error "Image: http://imgur.com/$FID"
 echo "$DEL_HASH"
}
image_delete_imgur() {
 local -r HID=$1
 local -r BASE_API='https://api.imgur.com/3'
 local RESPONSE
 log_debug 'deleting image from Imgur.com'
 RESPONSE=$(curl -X DELETE \
  -H 'Authorization: Client-ID 5926b561daf4510' \
  "$BASE_API/image/$HID.json") || return
 if ! match_json_true 'success' "$RESPONSE"; then
  local MSG ERRNO
  MSG=$(parse_json error <<< "$RESPONSE")
  ERRNO=$(parse_json status <<< "$RESPONSE")
  log_notice "$FUNCNAME: remote error, $MSG ($ERRNO)"
 fi
}
log_notice_stack() {
 local N
 for N in "${!FUNCNAME[@]}"; do
  [ $N -le 1 ] && continue
  log_notice "Failed inside ${FUNCNAME[$N]}(), line ${BASH_LINENO[$((N-1))]}, $(basename_file "${BASH_SOURCE[$N]}")"
  [[ ${BASH_SOURCE[$N]} = */core.sh ]] || break
 done
}
log_notice_norc() {
 if [[ $COLOR ]]; then
  test $VERBOSE -lt 2 || echo -ne "\033[0;33m$@\033[0m" >&2
 else
  test $VERBOSE -lt 2 || stderr -ne "$@"
 fi
}
command_not_found_handle() {
 local -r CMD=$1
 local ERR=$ERR_SYSTEM
 if [[ $CMD =~ _(delete|download|list|probe|upload)$ ]]; then
  stderr "$MODULE module: \`$CMD' function was not found"
 else
  [ "$CMD" = 'curl' ] && ERR=62
  stderr "$CMD: command not found"
 fi
 shift
 stderr "called with arguments: $*"
 return $ERR
}
